from typing import Callable, Dict, Optional, cast

from graphql.type import GraphQLNamedType, GraphQLObjectType, GraphQLSchema

from .resolvers import resolve_to
from .types import Resolver, SchemaBindable


class ObjectType(SchemaBindable):
    """Bindable populating object types in a GraphQL schema with Python logic.

    # Example

    Following code creates a GraphQL schema with single object type named `Query`
    and uses `ObjectType` to set resolvers on its fields:

    ```python
    import random
    from datetime import datetime

    from ariadne import ObjectType, make_executable_schema

    query_type = ObjectType("Query")

    @query_type.field("diceRoll")
    def resolve_dice_roll(*_):
        return random.int(1, 6)


    @query_type.field("year")
    def resolve_year(*_):
        return datetime.today().year


    schema = make_executable_schema(
        \"\"\"
        type Query {
            diceRoll: Int!
            year: Int!
        }
        \"\"\",
        query_type,
    )
    ```


    # Example with objects in objects

    When a field in the schema returns other GraphQL object, this object's
    resolvers are called with value returned from field's resolver. For example
    if there's an `user` field on the `Query` type that returns the `User` type,
    you don't have to resolve `User` fields in `user` resolver. In below example
    `fullName` field on `User` type is resolved from data on `UserModel` object
    that `user` field resolver on `Query` type returned:

    ```python
    import dataclasses
    from ariadne import ObjectType, make_executable_schema

    @dataclasses.dataclass
    class UserModel:
        id: int
        username: str
        first_name: str
        last_name: str


    users = [
        UserModel(
            id=1,
            username="Dany",
            first_name="Daenerys",
            last_name="Targaryen",
        ),
        UserModel(
            id=2,
            username="BlackKnight19",
            first_name="Cahir",
            last_name="Mawr Dyffryn aep Ceallach",
        ),
        UserModel(
            id=3,
            username="TheLady",
            first_name="Dorotea",
            last_name="Senjak",
        ),
    ]


    # Query type resolvers return users, but don't care about fields
    # of User type
    query_type = ObjectType("Query")

    @query_type.field("users")
    def resolve_users(*_) -> list[UserModel]:
        # In real world applications this would be a database query
        # returning iterable with user results
        return users


    @query_type.field("user")
    def resolve_user(*_, id: str) -> UserModel | None:
        # In real world applications this would be a database query
        # returning single user or None

        try:
            # GraphQL ids are always strings
            clean_id = int(id)
        except (ValueError, TypeError):
            # We could raise "ValueError" instead
            return None

        for user in users:
            if user.id == id:
                return user

        return None


    # User type resolvers don't know how to retrieve User, but know how to
    # resolve User type fields from UserModel instance
    user_type = ObjectType("User")

    # Resolve "name" GraphQL field to "username" attribute
    user_type.set_alias("name", "username")

    # Resolve "fullName" field to combined first and last name
    # `obj` argument will be populated by GraphQL with a value from
    # resolver for field returning "User" type
    @user_type.field("fullName")
    def resolve_user_full_name(obj: UserModel, *_):
        return f"{obj.first_name} {obj.last_name}"


    schema = make_executable_schema(
        \"\"\"
        type Query {
            users: [User!]!
            user(id: ID!): User
        }

        type User {
            id: ID!
            name: String!
            fullName: String!
        }
        \"\"\",
        query_type,
        user_type,
    )
    ```
    """

    _resolvers: Dict[str, Resolver]

    def __init__(self, name: str) -> None:
        """Initializes the `ObjectType` with a `name`.

        # Required arguments

        `name`: a `str` with the name of GraphQL object type in GraphQL schema to
        bind to.
        """
        self.name = name
        self._resolvers = {}

    def field(self, name: str) -> Callable[[Resolver], Resolver]:
        """Return a decorator that sets decorated function as a resolver for named field.

        Wrapper for `create_register_resolver` that on runtime validates `name` to be a
        string.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        bind decorated resolver to.
        """
        if not isinstance(name, str):
            raise ValueError(
                'field decorator should be passed a field name: @foo.field("name")'
            )
        return self.create_register_resolver(name)

    def create_register_resolver(self, name: str) -> Callable[[Resolver], Resolver]:
        """Return a decorator that sets decorated function as a resolver for named field.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        bind decorated resolver to.
        """

        def register_resolver(f: Resolver) -> Resolver:
            self._resolvers[name] = f
            return f

        return register_resolver

    def set_field(self, name, resolver: Resolver) -> Resolver:
        """Set a resolver for the field name.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        set this resolver for.

        `resolver`: a `Resolver` function to use.
        """
        self._resolvers[name] = resolver
        return resolver

    def set_alias(self, name: str, to: str) -> None:
        """Set an alias resolver for the field name to given Python nme.

        # Required arguments

        `name`: a `str` with a name of the GraphQL object's field in GraphQL schema to
        set this resolver for.

        `to`: a `str` of an attribute or dict key to resolve this field to.
        """
        self._resolvers[name] = resolve_to(to)

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Binds this `ObjectType` instance to the instance of GraphQL schema.

        If it has any resolver functions set, it assigns those to GraphQL type's
        fields `resolve` attributes. If field already has other resolver set on
        its `resolve` attribute, this resolver is replaced with the new one.
        """
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)
        graphql_type = cast(GraphQLObjectType, graphql_type)
        self.bind_resolvers_to_graphql_type(graphql_type)

    def validate_graphql_type(self, graphql_type: Optional[GraphQLNamedType]) -> None:
        """Validates that schema's GraphQL type associated with this `ObjectType`
        is a `type`."""
        if not graphql_type:
            raise ValueError("Type %s is not defined in the schema" % self.name)
        if not isinstance(graphql_type, GraphQLObjectType):
            raise ValueError(
                "%s is defined in the schema, but it is instance of %s (expected %s)"
                % (self.name, type(graphql_type).__name__, GraphQLObjectType.__name__)
            )

    def bind_resolvers_to_graphql_type(self, graphql_type, replace_existing=True):
        """Binds this `ObjectType` instance to the instance of GraphQL schema."""
        for field, resolver in self._resolvers.items():
            if field not in graphql_type.fields:
                raise ValueError(
                    "Field %s is not defined on type %s" % (field, self.name)
                )
            if graphql_type.fields[field].resolve is None or replace_existing:
                graphql_type.fields[field].resolve = resolver


class QueryType(ObjectType):
    """An convenience class for defining Query type.

    # Example

    Both of those code samples have same effect:

    ```python
    query_type = QueryType()
    ```

    ```python
    query_type = ObjectType("Query")
    ```
    """

    def __init__(self) -> None:
        """Initializes the `QueryType` with a GraphQL name set to `Query`."""
        super().__init__("Query")


class MutationType(ObjectType):
    """An convenience class for defining Mutation type.

    # Example

    Both of those code samples have same result:

    ```python
    mutation_type = MutationType()
    ```

    ```python
    mutation_type = ObjectType("Mutation")
    ```
    """

    def __init__(self) -> None:
        """Initializes the `MutationType` with a GraphQL name set to `Mutation`."""
        super().__init__("Mutation")



from contextlib import contextmanager
from typing import List, Optional, Type

from graphql import GraphQLError
from graphql.execution import MiddlewareManager

from .types import MiddlewareList, ContextValue, ExtensionList


class ExtensionManager:
    """Container and runner for extensions and middleware, used by the GraphQL servers.

    # Attributes

    `context`: the `ContextValue` of type specific to the server.

    `extensions`: a `tuple` with instances of initialized extensions.

    `extensions_reversed`: a `tuple` created from reversing `extensions`.
    """

    __slots__ = ("context", "extensions", "extensions_reversed")

    def __init__(
        self,
        extensions: Optional[ExtensionList] = None,
        context: Optional[ContextValue] = None,
    ) -> None:
        """Initializes extensions and stores them with context on instance.

        # Optional arguments

        `extensions`: a `list` of `Extension` types to initialize.

        `context`: the `ContextValue` of type specific to the server.
        """
        self.context = context

        if extensions:
            self.extensions = tuple(ext() for ext in extensions)
            self.extensions_reversed = tuple(reversed(self.extensions))
        else:
            self.extensions_reversed = self.extensions = tuple()

    def as_middleware_manager(
        self,
        middleware: MiddlewareList = None,
        manager_class: Optional[Type[MiddlewareManager]] = None,
    ) -> Optional[MiddlewareManager]:
        """Creates middleware manager instance combining middleware and extensions.

        Returns instance of the type passed in `manager_class` argument
        or `MiddlewareManager` that query executor then uses.

        # Optional arguments

        `middleware`: a `list` of `Middleware` instances

        `manager_class` a `type` of middleware manager to use. `MiddlewareManager`
        is used if this argument is passed `None` or omitted.
        """
        if not middleware and not self.extensions:
            return None

        middleware = middleware or []
        if manager_class:
            return manager_class(*middleware, *self.extensions)

        return MiddlewareManager(*middleware, *self.extensions)

    @contextmanager
    def request(self):
        """A context manager that should wrap request processing.

        Runs `request_started` hook at beginning and `request_finished` at
        the end of request processing, enabling APM extensions like ApolloTracing.
        """
        for ext in self.extensions:
            ext.request_started(self.context)
        try:
            yield
        finally:
            for ext in self.extensions_reversed:
                ext.request_finished(self.context)

    def has_errors(self, errors: List[GraphQLError]):
        """Propagates GraphQL errors returned by GraphQL server to extensions.

        Should be called only when there are errors.
        """
        for ext in self.extensions:
            ext.has_errors(errors, self.context)

    def format(self) -> dict:
        """Gathers data from extensions for inclusion in server's response JSON.

        This data can be retrieved from the `extensions` key in response JSON.

        Returns `dict` with JSON-serializable data.
        """
        data = {}
        for ext in self.extensions:
            ext_data = ext.format(self.context)
            if ext_data:
                data.update(ext_data)
        return data



from reprlib import repr  # pylint: disable=redefined-builtin
from traceback import format_exception

from typing import List, Optional, cast

from graphql import GraphQLError

from .utils import unwrap_graphql_error


def format_error(error: GraphQLError, debug: bool = False) -> dict:
    """Format the GraphQL error into JSON serializable format.

    If `debug` is set to `True`, error's JSON will also include the `extensions`
    key with `exception` object containing error's `context` and `stacktrace`.

    Returns a JSON-serializable `dict` with error representation.

    # Required arguments

    `error`: an `GraphQLError` to convert into JSON serializable format.

    # Optional arguments

    `debug`: a `bool` that controls if debug data should be included in
    result `dict`. Defaults to `False`.
    """

    formatted = cast(dict, error.formatted)
    if debug:
        if "extensions" not in formatted:
            formatted["extensions"] = {}
        formatted["extensions"]["exception"] = get_error_extension(error)
    return formatted


def get_error_extension(error: GraphQLError) -> Optional[dict]:
    """Get a JSON-serializable `dict` containing error's stacktrace and context.

    Returns a JSON-serializable `dict` with `stacktrace` and `context` to include
    under error's `extensions` key in JSON response. Returns `None` if `error`
    has no stacktrace or wraps no exception.

    # Required arguments

    `error`: an `GraphQLError` to return context and stacktrace for.
    """
    unwrapped_error = unwrap_graphql_error(error)
    if unwrapped_error is None or not error.__traceback__:
        return None

    unwrapped_error = cast(Exception, unwrapped_error)
    return {
        "stacktrace": get_formatted_error_traceback(unwrapped_error),
        "context": get_formatted_error_context(unwrapped_error),
    }


def get_formatted_error_traceback(error: Exception) -> List[str]:
    """Get JSON-serializable stacktrace from `Exception`.

    Returns list of strings, with every item being separate line from stacktrace.

    This approach produces better results in GraphQL explorers which display every
    line under previous one but not always format linebreak characters for blocks
    of text.

    # Required arguments

    `error`: an `Exception` to return formatted stacktrace for.
    """
    formatted = []
    for line in format_exception(type(error), error, error.__traceback__):
        formatted.extend(line.rstrip().splitlines())
    return formatted


def get_formatted_error_context(error: Exception) -> Optional[dict]:
    """Get JSON-serializable context from `Exception`.

    Returns a `dict` of strings, with every key being value name and value
    being `repr()` of it's Python value. Returns `None` if context is not
    available.

    # Required arguments

    `error`: an `Exception` to return formatted context for.
    """

    tb_last = error.__traceback__
    while tb_last and tb_last.tb_next:
        tb_last = tb_last.tb_next
    if tb_last is None:
        return None
    return {key: repr(value) for key, value in tb_last.tb_frame.f_locals.items()}



from enum import Enum
from typing import Dict, List, Optional, Type, Union

from graphql import (
    GraphQLSchema,
    assert_valid_schema,
    build_ast_schema,
    parse,
)

from .enums import (
    EnumType,
    set_default_enum_values_on_schema,
    validate_schema_enum_values,
)
from .schema_names import SchemaNameConverter, convert_schema_names
from .schema_visitor import SchemaDirectiveVisitor
from .types import SchemaBindable

SchemaBindables = Union[
    SchemaBindable,
    Type[Enum],
    List[Union[SchemaBindable, Type[Enum]]],
]


def make_executable_schema(
    type_defs: Union[str, List[str]],
    *bindables: SchemaBindables,
    directives: Optional[Dict[str, Type[SchemaDirectiveVisitor]]] = None,
    convert_names_case: Union[bool, SchemaNameConverter] = False,
) -> GraphQLSchema:
    """Create a `GraphQLSchema` instance that can be used to execute queries.

    Returns a `GraphQLSchema` instance with attributes populated with Python
    values and functions.

    # Required arguments

    `type_defs`: a `str` or list of `str` with GraphQL types definitions in
    schema definition language (`SDL`).

    # Optional arguments

    `bindables`: instances or lists of instances of schema bindables. Order in
    which bindables are passed to `make_executable_schema` matters depending on
    individual bindable's implementation.

    `directives`: a dict of GraphQL directives to apply to schema. Dict's keys must
    correspond to directives names in GraphQL schema and values should be
    `SchemaDirectiveVisitor` classes (_not_ instances) implementing their logic.

    `convert_names_case`: a `bool` or function of `SchemaNameConverter` type to
    use to convert names in GraphQL schema between `camelCase` used by GraphQL
    and `snake_case` used by Python. Defaults to `False`, making all conversion
    explicit and up to developer to implement. Set `True` to use
    default strategy using `convert_camel_case_to_snake` for name conversions or
    set to custom function to customize this behavior.

    # Example with minimal schema

    Below code creates minimal executable schema that doesn't implement any Python
    logic, but still executes queries using `root_value`:

    ```python
    from ariadne import graphql_sync, make_executable_schema

    schema = make_executable_schema(
        \"\"\"
        type Query {
            helloWorld: String!
        }
        \"\"\"
    )

    no_errors, result = graphql_sync(
        schema,
        {"query": "{ helloWorld }"},
        root_value={"helloWorld": "Hello world!"},
    )

    assert no_errors
    assert result == {
        "data": {
            "helloWorld": "Hello world!",
        },
    }
    ```

    # Example with bindables

    Below code creates executable schema that combines different ways of passing
    bindables to add Python logic to schema:

    ```python
    from dataclasses import dataclass
    from enum import Enum
    from ariadne import ObjectType, QueryType, UnionType, graphql_sync, make_executable_schema

    # Define some types representing database models in real applications
    class UserLevel(str, Enum):
        USER = "user"
        ADMIN = "admin"

    @dataclass
    class UserModel:
        id: str
        name: str
        level: UserLevel

    @dataclass
    class PostModel:
        id: str
        body: str

    # Create fake "database"
    results = (
        UserModel(id=1, name="Bob", level=UserLevel.USER),
        UserModel(id=2, name="Alice", level=UserLevel.ADMIN),
        UserModel(id=3, name="Jon", level=UserLevel.USER),
        PostModel(id=1, body="Hello world!"),
        PostModel(id=2, body="How's going?"),
        PostModel(id=3, body="Sure thing!"),
    )


    # Resolve username field in GraphQL schema to user.name attribute
    user_type = ObjectType("User")
    user_type.set_alias("username", "name")


    # Resolve message field in GraphQL schema to post.body attribute
    post_type = ObjectType("Post")
    post_type.set_alias("message", "body")


    # Resolve results field in GraphQL schema to results array
    query_type = QueryType()

    @query_type.field("results")
    def resolve_results(*_):
        return results


    # Resolve GraphQL type of individual result from it's Python class
    result_type = UnionType("Result")

    @result_type.type_resolver
    def resolve_result_type(obj: UserModel | PostModel | dict, *_) -> str:
        if isinstance(obj, UserModel):
            return "User"

        if isinstance(obj, PostModel):
            return "Post"

        raise ValueError(f"Don't know GraphQL type for '{obj}'!")


    # Create executable schema that returns list of results
    schema = make_executable_schema(
        \"\"\"
        type Query {
            results: [Result!]!
        }

        union Result = User | Post

        type User {
            id: ID!
            username: String!
            level: UserLevel!
        }

        type Post {
            id: ID!
            message: String!
        }

        enum UserLevel {
            USER
            ADMIN
        }
        \"\"\",
        # Bindables *args accept single instances:
        query_type,
        result_type,
        # Bindables *args accepts lists of instances:
        [user_type, post_type],
        # Both approaches can be mixed
        # Python Enums are also valid bindables:
        UserLevel,
    )

    # Query the schema for results
    no_errors, result = graphql_sync(
        schema,
        {
            "query": (
                \"\"\"
                {
                    results {
                        ... on Post {
                            id
                            message
                        }
                        ... on User {
                            id
                            username
                            level
                        }
                    }
                }
                \"\"\"
            ),
        },
    )

    # Verify that it works
    assert no_errors
    assert result == {
        "data": {
            "results": [
                {
                    "id": "1",
                    "username": "Bob",
                    "level": "USER",
                },
                {
                    "id": "2",
                    "username": "Alice",
                    "level": "ADMIN",
                },
                {
                    "id": "3",
                    "username": "Jon",
                    "level": "USER",
                },
                {
                    "id": "1",
                    "message": "Hello world!",
                },
                {
                    "id": "2",
                    "message": "How's going?",
                },
                {
                    "id": "3",
                    "message": "Sure thing!",
                },
            ],
        },
    }
    ```

    # Example with directive

    Below code uses `directives` option to set custom directive on schema:

    ```python
    from functools import wraps
    from ariadne import SchemaDirectiveVisitor, graphql_sync, make_executable_schema
    from graphql import default_field_resolver

    class UppercaseDirective(SchemaDirectiveVisitor):
        def visit_field_definition(self, field, object_type):
            org_resolver = field.resolve or default_field_resolver

            @wraps(org_resolver)
            def uppercase_resolved_value(*args, **kwargs):
                value = org_resolver(*args, **kwargs)
                if isinstance(value, str):
                    return value.upper()
                return value

            # Extend field's behavior by wrapping it's resolver in custom one
            field.resolve = uppercase_resolved_value
            return field


    schema = make_executable_schema(
        \"\"\"
        directive @uppercase on FIELD_DEFINITION

        type Query {
            helloWorld: String! @uppercase
        }
        \"\"\",
        directives={"uppercase": UppercaseDirective},
    )

    no_errors, result = graphql_sync(
        schema,
        {"query": "{ helloWorld }"},
        root_value={"helloWorld": "Hello world!"},
    )

    assert no_errors
    assert result == {
        "data": {
            "helloWorld": "HELLO WORLD!",
        },
    }
    ```

    # Example with converted names

    Below code uses `convert_names_case=True` option to resolve `helloWorld`
    field to `hello_world` key from `root_value`:

    ```python
    from ariadne import graphql_sync, make_executable_schema

    schema = make_executable_schema(
        \"\"\"
        type Query {
            helloWorld: String!
        }
        \"\"\",
        convert_names_case=True,
    )

    no_errors, result = graphql_sync(
        schema,
        {"query": "{ helloWorld }"},
        root_value={"hello_world": "Hello world!"},
    )

    assert no_errors
    assert result == {
        "data": {
            "helloWorld": "Hello world!",
        },
    }
    ```
    """

    if isinstance(type_defs, list):
        type_defs = join_type_defs(type_defs)

    ast_document = parse(type_defs)
    schema = build_ast_schema(ast_document)
    normalized_bindables = normalize_bindables(*bindables)

    for bindable in normalized_bindables:
        if isinstance(bindable, SchemaBindable):
            bindable.bind_to_schema(schema)

    set_default_enum_values_on_schema(schema)

    if directives:
        SchemaDirectiveVisitor.visit_schema_directives(schema, directives)

    assert_valid_schema(schema)
    validate_schema_enum_values(schema)
    repair_default_enum_values(schema, normalized_bindables)

    if convert_names_case:
        convert_schema_names(
            schema,
            convert_names_case if callable(convert_names_case) else None,
        )

    return schema


def join_type_defs(type_defs: List[str]) -> str:
    return "\n\n".join(t.strip() for t in type_defs)


def normalize_bindables(
    *bindables: SchemaBindables,
) -> List[SchemaBindable]:
    normal_bindables: List[SchemaBindable] = []
    for bindable in flatten_bindables(*bindables):
        if isinstance(bindable, SchemaBindable):
            normal_bindables.append(bindable)
        elif issubclass(bindable, Enum):
            normal_bindables.append(EnumType(bindable.__name__, bindable))
        else:
            raise ValueError(f"Unsupported type: {repr(bindable)}")

    return normal_bindables


def flatten_bindables(
    *bindables: SchemaBindables,
) -> List[Union[SchemaBindable, Type[Enum]]]:
    new_bindables = []

    for bindable in bindables:
        if isinstance(bindable, list):
            new_bindables.extend(bindable)
        else:
            new_bindables.append(bindable)

    return new_bindables


def repair_default_enum_values(
    schema: GraphQLSchema, bindables: List[SchemaBindable]
) -> None:
    for bindable in bindables:
        if isinstance(bindable, EnumType):
            bindable.bind_to_default_values(schema)



from types import FunctionType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from graphql import is_named_type, value_from_ast_untyped
from graphql.execution.values import get_argument_values
from graphql.language import DirectiveLocation
from graphql.type import (
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNamedType,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
    GraphQLUnionType,
)
from typing_extensions import Literal, Protocol

VisitableSchemaType = Union[
    GraphQLSchema,
    GraphQLObjectType,
    GraphQLInterfaceType,
    GraphQLInputObjectType,
    GraphQLNamedType,
    GraphQLScalarType,
    GraphQLField,
    GraphQLArgument,
    GraphQLUnionType,
    GraphQLEnumType,
    GraphQLEnumValue,
]
V = TypeVar("V", bound=VisitableSchemaType)
VisitableMap = Dict[str, V]
IndexedObject = Union[VisitableMap, Tuple[V, ...]]


Callback = Callable[..., Any]


def each(tuple_or_dict: IndexedObject, callback: Callback):
    if isinstance(tuple_or_dict, tuple):
        for value in tuple_or_dict:
            callback(value)
    else:
        for key, value in tuple_or_dict.items():
            callback(value, key)


def update_each_key(object_map: VisitableMap, callback: Callback):
    """
    The callback can return None to leave the key untouched, False to remove
    the key from the array or object, or a non-null V to replace the value.
    """

    keys_to_remove: List[str] = []

    for key, value in object_map.copy().items():
        result = callback(value, key)
        if result is False:
            keys_to_remove.append(key)
        elif result is None:
            continue
        else:
            object_map[key] = result
    for k in keys_to_remove:
        object_map.pop(k)


class SchemaVisitor(Protocol):
    @classmethod
    def implements_visitor_method(cls, method_name: str):
        if not method_name.startswith("visit_"):
            return False

        try:
            method = getattr(cls, method_name)
        except AttributeError:
            return False

        if not isinstance(method, FunctionType):
            return False

        if cls.__name__ == "SchemaVisitor":
            # The SchemaVisitor class implements every visitor method.
            return True

        if method.__qualname__.startswith("SchemaVisitor"):
            #  When SchemaVisitor subclass does not really implement the method.
            return False

        return True

    # pylint: disable=unused-argument
    def visit_schema(self, schema: GraphQLSchema) -> None:
        pass

    def visit_scalar(self, scalar: GraphQLScalarType) -> GraphQLScalarType:
        pass

    def visit_object(self, object_: GraphQLObjectType) -> GraphQLObjectType:
        pass

    def visit_field_definition(
        self,
        field: GraphQLField,
        object_type: Union[GraphQLObjectType, GraphQLInterfaceType],
    ) -> GraphQLField:
        pass

    def visit_argument_definition(
        self,
        argument: GraphQLArgument,
        field: GraphQLField,
        object_type: Union[GraphQLObjectType, GraphQLInterfaceType],
    ) -> GraphQLArgument:
        pass

    def visit_interface(self, interface: GraphQLInterfaceType) -> GraphQLInterfaceType:
        pass

    def visit_union(self, union: GraphQLUnionType) -> GraphQLUnionType:
        pass

    def visit_enum(self, type_: GraphQLEnumType) -> GraphQLEnumType:
        pass

    def visit_enum_value(
        self, value: GraphQLEnumValue, enum_type: GraphQLEnumType
    ) -> GraphQLEnumValue:
        pass

    def visit_input_object(
        self, object_: GraphQLInputObjectType
    ) -> GraphQLInputObjectType:
        pass

    def visit_input_field_definition(
        self, field: GraphQLInputField, object_type: GraphQLInputObjectType
    ) -> GraphQLInputField:
        pass


def visit_schema(
    schema: GraphQLSchema,
    visitor_selector: Callable[
        [VisitableSchemaType, str], List["SchemaDirectiveVisitor"]
    ],
) -> GraphQLSchema:
    """
    Helper function that calls visitor_selector and applies the resulting
    visitors to the given type, with arguments [type, ...args].
    """

    def call_method(
        method_name: str, type_: VisitableSchemaType, *args: Any
    ) -> Union[VisitableSchemaType, Literal[False]]:
        for visitor in visitor_selector(type_, method_name):
            new_type = getattr(visitor, method_name)(type_, *args)
            if new_type is None:
                # Keep going without modifying type.
                continue

            if method_name == "visit_schema" or isinstance(type_, GraphQLSchema):
                raise ValueError(
                    f"Method {method_name} cannot replace schema with {new_type}"
                )

            if new_type is False:
                # Stop the loop and return False form call_method, which will cause
                # the type to be removed from the schema.
                del type_
                return False

            # Update type to the new type returned by the visitor method, so that
            # later directives will see the new type, and call_method will return
            # the final type.
            type_ = new_type

        # If there were no directives for this type object, or if all visitor
        # methods returned nothing, type will be returned unmodified.
        return type_

    def visit(  # pylint: disable=too-many-return-statements
        type_: VisitableSchemaType,
    ) -> Union[VisitableSchemaType, Literal[False]]:
        """
        Recursive helper function that calls any appropriate visitor methods for
        each object in the schema, then traverses the object's children (if any).
        """
        if isinstance(type_, GraphQLSchema):
            # Unlike the other types, the root GraphQLSchema object cannot be
            # replaced by visitor methods, because that would make life very hard
            # for SchemaVisitor subclasses that rely on the original schema object.
            call_method("visit_schema", type_)

            def _start(named_type, type_name):
                if not type_name.startswith("__"):
                    visit(named_type)

            update_each_key(type_.type_map, _start)

            return type_

        if isinstance(type_, GraphQLObjectType):
            # Note that call_method('visit_object', type_) may not actually call any
            # methods, if there are no @directive annotations associated with this
            # type, or if this SchemaDirectiveVisitor subclass does not override
            # the visit_object method.
            new_object = cast(GraphQLObjectType, call_method("visit_object", type_))
            if new_object:
                visit_fields(new_object)

            return new_object

        if isinstance(type_, GraphQLInterfaceType):
            new_interface = cast(
                GraphQLInterfaceType, call_method("visit_interface", type_)
            )
            if new_interface:
                visit_fields(new_interface)

            return new_interface

        if isinstance(type_, GraphQLInputObjectType):
            new_input_object = cast(
                GraphQLInputObjectType, call_method("visit_input_object", type_)
            )

            if new_input_object:
                update_each_key(
                    new_input_object.fields,
                    lambda field, n: call_method(
                        "visit_input_field_definition", field, new_input_object
                    ),
                )

            return new_input_object

        if isinstance(type_, GraphQLScalarType):
            return call_method("visit_scalar", type_)

        if isinstance(type_, GraphQLUnionType):
            return call_method("visit_union", type_)

        if isinstance(type_, GraphQLEnumType):
            new_enum = cast(GraphQLEnumType, call_method("visit_enum", type_))

            if new_enum:
                update_each_key(
                    new_enum.values,
                    lambda value, name: call_method("visit_enum_value", value, name),
                )

            return new_enum

        raise ValueError(f"Unexpected schema type: {type_}")

    def visit_fields(type_: Union[GraphQLObjectType, GraphQLInterfaceType]):
        def _update_fields(field, _):
            # It would be nice if we could call visit(field) recursively here, but
            # GraphQLField is merely a type, not a value that can be detected using
            # an instanceof check, so we have to visit the fields in this lexical
            # context, so that TypeScript can validate the call to
            # visit_field_definition.
            new_field = cast(
                GraphQLField, call_method("visit_field_definition", field, type_)
            )
            # While any field visitor needs a reference to the field object, some
            # field visitors may also need to know the enclosing (parent) type,
            # perhaps to determine if the parent is a GraphQLObjectType or a
            # GraphQLInterfaceType. To obtain a reference to the parent, a
            # visitor method can have a second parameter, which will be referring
            # to the parent.

            if new_field and new_field.args:
                update_each_key(
                    new_field.args,
                    lambda arg, _: call_method(
                        "visit_argument_definition", arg, new_field, type_
                    ),
                )

            return new_field

        update_each_key(type_.fields, _update_fields)

    visit(schema)

    # Return the original schema for convenience, even though it cannot have
    # been replaced or removed by the code above.
    return schema


def directive_location_to_visitor_method_name(loc: DirectiveLocation):
    """
    Convert a string like "FIELD_DEFINITION" to "visit_field_definition".
    """
    return "visit_" + loc.name.lower()


class SchemaDirectiveVisitor(SchemaVisitor):
    """Base class for custom GraphQL directives.

    Also implements class methods with container and management logic for
    directives at schema creation time, used by `make_executable_schema`.

    # Lifecycle

    Separate instances of the GraphQL directive are created for each GraphQL
    schema item with the directive set on it. If directive is set on two
    fields, two separate instances of a directive will be created.

    # Example schema visitors

    `SchemaDirectiveVisitor` subclasses can implement any of below methods
    that will be called when directive is applied to different elements of
    GraphQL schema:

    ```python
    from ariadne import SchemaDirectiveVisitor
    from graphql import (
        GraphQLArgument,
        GraphQLEnumType,
        GraphQLEnumValue,
        GraphQLField,
        GraphQLInputField,
        GraphQLInputObjectType,
        GraphQLInterfaceType,
        GraphQLObjectType,
        GraphQLScalarType,
        GraphQLSchema,
        GraphQLUnionType,
    )

    class MyDirective(SchemaDirectiveVisitor):
        def visit_schema(self, schema: GraphQLSchema) -> None:
            pass

        def visit_scalar(self, scalar: GraphQLScalarType) -> GraphQLScalarType:
            pass

        def visit_object(self, object_: GraphQLObjectType) -> GraphQLObjectType:
            pass

        def visit_field_definition(
            self,
            field: GraphQLField,
            object_type: Union[GraphQLObjectType, GraphQLInterfaceType],
        ) -> GraphQLField:
            pass

        def visit_argument_definition(
            self,
            argument: GraphQLArgument,
            field: GraphQLField,
            object_type: Union[GraphQLObjectType, GraphQLInterfaceType],
        ) -> GraphQLArgument:
            pass

        def visit_interface(self, interface: GraphQLInterfaceType) -> GraphQLInterfaceType:
            pass

        def visit_union(self, union: GraphQLUnionType) -> GraphQLUnionType:
            pass

        def visit_enum(self, type_: GraphQLEnumType) -> GraphQLEnumType:
            pass

        def visit_enum_value(
            self, value: GraphQLEnumValue, enum_type: GraphQLEnumType
        ) -> GraphQLEnumValue:
            pass

        def visit_input_object(
            self, object_: GraphQLInputObjectType
        ) -> GraphQLInputObjectType:
            pass

        def visit_input_field_definition(
            self, field: GraphQLInputField, object_type: GraphQLInputObjectType
        ) -> GraphQLInputField:
            pass
    ```
    """

    def __init__(self, name, args, visited_type, schema, context) -> None:
        """Instantiates the directive for schema object.

        # Required arguments

        `name`: a `str` with directive's name.

        `args`: a `dict` with directive's arguments.

        `visited_type`: an GraphQL type this directive is set on.

        `schema`: the GraphQL schema instance.

        `context`: `None`, unused but present for historic reasons.
        """

        self.name = name
        self.args = args
        self.visited_type = visited_type
        self.schema = schema
        self.context = context

    @classmethod
    def get_directive_declaration(
        cls, directive_name: str, schema: GraphQLSchema
    ) -> Optional[GraphQLDirective]:
        """Get GraphQL directive declaration from GraphQL schema by it's name.

        Returns `GraphQLDirective` object or `None`.

        # Required arguments

        `directive_name`: a `str` with name of directive in the GraphQL schema.

        `schema`: a `GraphQLSchema` instance to retrieve the directive
        declaration from.
        """
        return schema.get_directive(directive_name)

    @classmethod
    def get_declared_directives(
        cls,
        schema: GraphQLSchema,
        directive_visitors: Dict[str, Type["SchemaDirectiveVisitor"]],
    ) -> Dict[str, GraphQLDirective]:
        """Get GraphQL directives declaration from GraphQL schema by their names.

        Returns a `dict` where keys are strings with directive names in schema
        and values are `GraphQLDirective` objects with their declarations in the
        GraphQL schema.

        # Required arguments

        `directive_name`: a `str` with name of directive in the GraphQL schema.

        `schema`: a `GraphQLSchema` instance to retrieve the directive
        declaration from.
        """

        declared_directives: Dict[str, GraphQLDirective] = {}

        def _add_directive(decl):
            declared_directives[decl.name] = decl

        each(schema.directives, _add_directive)

        #  If the visitor subclass overrides get_directive_declaration, and it
        #  returns a non-null GraphQLDirective, use that instead of any directive
        #  declared in the schema itself. Reasoning: if a SchemaDirectiveVisitor
        #  goes to the trouble of implementing get_directive_declaration, it should
        #  be able to rely on that implementation.
        def _get_overriden_directive(visitor_class, directive_name):
            decl = visitor_class.get_directive_declaration(directive_name, schema)
            if decl:
                declared_directives[directive_name] = decl

        each(directive_visitors, _get_overriden_directive)

        def _rest(decl, name):
            if not name in directive_visitors:
                #  SchemaDirectiveVisitors.visit_schema_directives might be called
                #  multiple times with partial directive_visitors maps, so it's not
                #  necessarily an error for directive_visitors to be missing an
                #  implementation of a directive that was declared in the schema.
                return

            visitor_class = directive_visitors[name]

            def _location_check(loc):
                visitor_method_name = directive_location_to_visitor_method_name(loc)

                if SchemaVisitor.implements_visitor_method(
                    visitor_method_name
                ) and not visitor_class.implements_visitor_method(visitor_method_name):
                    #  While visitor subclasses may implement extra visitor methods,
                    #  it's definitely a mistake if the GraphQLDirective declares itself
                    #  applicable to certain schema locations, and the visitor subclass
                    #  does not implement all the corresponding methods.
                    raise ValueError(
                        f"SchemaDirectiveVisitor for @{name} must"
                        f"implement {visitor_method_name} method"
                    )

            each(decl.locations, _location_check)

        each(declared_directives, _rest)

        return declared_directives

    @classmethod
    def visit_schema_directives(
        cls,
        schema: GraphQLSchema,
        directive_visitors: Dict[str, Type["SchemaDirectiveVisitor"]],
        *,
        context: Optional[Dict[str, Any]] = None,
    ) -> Mapping[str, List["SchemaDirectiveVisitor"]]:
        """Apply directives to the GraphQL schema.

        Applied directives mutate the GraphQL schema in place.

        Returns dict with names of GraphQL directives as keys and list of
        directive instances created for each directive name.

        # Required arguments

        `schema`: a GraphQL schema to which directives should be applied.

        `directive_visitors`: a `dict` with `str` and
        `Type[SchemaDirectiveVisitor]` pairs defining mapping of
        `SchemaDirectiveVisitor` types to their names in the GraphQL schema.

        # Optional arguments

        `context`: `None`, unused but present for historic reasons.
        """
        declared_directives = cls.get_declared_directives(schema, directive_visitors)

        #  Map from directive names to lists of SchemaDirectiveVisitor instances
        #  created while visiting the schema.
        created_visitors: Dict[str, List["SchemaDirectiveVisitor"]] = {
            k: [] for k in directive_visitors
        }

        def _visitor_selector(
            type_: VisitableSchemaType, method_name: str
        ) -> List["SchemaDirectiveVisitor"]:
            visitors: List["SchemaDirectiveVisitor"] = []
            directive_nodes = type_.ast_node.directives if type_.ast_node else None
            if directive_nodes is None:
                return visitors

            for directive_node in directive_nodes:
                directive_name = directive_node.name.value
                if directive_name not in directive_visitors:
                    continue

                visitor_class = directive_visitors[directive_name]

                #  Avoid creating visitor objects if visitor_class does not override
                #  the visitor method named by method_name.
                if not visitor_class.implements_visitor_method(method_name):
                    continue

                decl = declared_directives[directive_name]

                args: Dict[str, Any] = {}

                if decl:
                    #  If this directive was explicitly declared, use the declared
                    #  argument types (and any default values) to check, coerce, and/or
                    #  supply default values for the given arguments.
                    args = get_argument_values(decl, directive_node)
                else:
                    #  If this directive was not explicitly declared, just convert the
                    #  argument nodes to their corresponding values.
                    for arg in directive_node.arguments:
                        args[arg.name.value] = value_from_ast_untyped(arg.value)

                #  As foretold in comments near the top of the visit_schema_directives
                #  method, this is where instances of the SchemaDirectiveVisitor class
                #  get created and assigned names. While subclasses could override the
                #  constructor method, the constructor is marked as protected, so
                #  these are the only arguments that will ever be passed.
                visitors.append(
                    visitor_class(directive_name, args, type_, schema, context)
                )

            for visitor in visitors:
                created_visitors[visitor.name].append(visitor)

            return visitors

        visit_schema(schema, _visitor_selector)

        # Automatically update any references to named schema types replaced
        # during the traversal, so implementors don't have to worry about that.
        heal_schema(schema)

        return created_visitors


NamedTypeMap = Dict[str, GraphQLNamedType]


def heal_schema(schema: GraphQLSchema) -> GraphQLSchema:
    def heal(type_: VisitableSchemaType):
        if isinstance(type_, GraphQLSchema):
            original_type_map: NamedTypeMap = type_.type_map
            actual_named_type_map: NamedTypeMap = {}

            def _heal_original(named_type, type_name):
                if type_name.startswith("__"):
                    return None

                actual_name = named_type.name
                if actual_name.startswith("__"):
                    return None

                if actual_name in actual_named_type_map:
                    raise ValueError(f"Duplicate schema type name {actual_name}")

                actual_named_type_map[actual_name] = named_type

                # Note: we are deliberately leaving named_type in the schema by its
                # original name (which might be different from actual_name), so that
                # references by that name can be healed.
                return None

            # If any of the .name properties of the GraphQLNamedType objects in
            # schema.type_map have changed, the keys of the type map need to
            # be updated accordingly.
            each(original_type_map, _heal_original)

            # Now add back every named type by its actual name.
            def _add_back(named_type, type_name):
                original_type_map[type_name] = named_type

            each(actual_named_type_map, _add_back)

            # Directive declaration argument types can refer to named types.
            def _heal_directive_declaration(decl: GraphQLDirective):
                def _heal_arg(arg, _):
                    arg.type = heal_type(arg.type)

                if decl.args:
                    each(decl.args, _heal_arg)

            each(type_.directives, _heal_directive_declaration)

            def _heal_type(named_type, type_name):
                if not type_name.startswith("__"):
                    heal(named_type)

            each(original_type_map, _heal_type)

            # Dangling references to renamed types should remain in the schema
            # during healing, but must be removed now, so that the following
            # invariant holds for all names: schema.get_type(name).name === name
            def _remove_dangling_references(_, type_name):
                if (
                    not type_name.startswith("__")
                    and type_name not in actual_named_type_map
                ):
                    return False
                return None

            update_each_key(original_type_map, _remove_dangling_references)

        elif isinstance(type_, GraphQLObjectType):
            heal_fields(type_)
            each(type_.interfaces, heal)

        elif isinstance(type_, GraphQLInterfaceType):
            heal_fields(type_)

        elif isinstance(type_, GraphQLInputObjectType):

            def _heal_field_type(field, _):
                field.type = heal_type(field.type)

            each(type_.fields, _heal_field_type)

        elif isinstance(type_, GraphQLScalarType):
            # Nothing to do.
            pass

        elif isinstance(type_, GraphQLUnionType):
            each(type_.types, heal_type)

        elif isinstance(type_, GraphQLEnumType):
            # Nothing to do.
            pass

        else:
            raise ValueError(f"Unexpected schema type: {type_}")

    def heal_fields(type_: Union[GraphQLObjectType, GraphQLInterfaceType]):
        def _heal_arg(arg, _):
            arg.type = heal_type(arg.type)

        def _heal_field(field, _):
            field.type = heal_type(field.type)
            if field.args:
                each(field.args, _heal_arg)

        each(type_.fields, _heal_field)

    def heal_type(
        type_: Union[GraphQLList, GraphQLNamedType, GraphQLNonNull]
    ) -> Union[GraphQLList, GraphQLNamedType, GraphQLNonNull]:
        # Unwrap the two known wrapper types
        if isinstance(type_, GraphQLList):
            type_ = GraphQLList(heal_type(type_.of_type))
        elif isinstance(type_, GraphQLNonNull):
            type_ = GraphQLNonNull(heal_type(type_.of_type))
        elif is_named_type(type_):
            # If a type annotation on a field or an argument or a union member is
            # any `GraphQLNamedType` with a `name`, then it must end up identical
            # to `schema.get_type(name)`, since `schema.type_map` is the source
            # of truth for all named schema types.
            named_type = cast(GraphQLNamedType, type_)
            official_type = schema.get_type(named_type.name)
            if official_type and named_type != official_type:
                return official_type

        return type_

    heal(schema)
    return schema



def escape_default_query(query: str, js_sign: str = "'") -> str:
    while "\\" + js_sign in query:
        query = query.replace("\\" + js_sign, js_sign)
    while "\\n" in query:
        query = query.replace("\\n", "\n")
    while "\\\n" in query:
        query = query.replace("\\\n", "\n")
    return query.replace("\n", "\\n").replace(js_sign, "\\" + js_sign)



from typing import Any, Awaitable, Optional, Union


class Explorer:
    def html(self, request: Any) -> Union[Optional[str], Awaitable[Optional[str]]]:
        raise NotImplementedError("Explorer subclasses should define 'html' method")


class ExplorerHttp405(Explorer):
    """Explorer that always returns HTTP 405 not allowed response."""

    def html(self, _):
        return None



from typing import Optional, cast

from graphql.type import (
    GraphQLInterfaceType,
    GraphQLNamedType,
    GraphQLSchema,
)

from .objects import ObjectType
from .types import Resolver
from .utils import type_implements_interface


class InterfaceType(ObjectType):
    """Bindable populating interfaces in a GraphQL schema with Python logic.

    Extends `ObjectType`, providing `field` decorator and `set_field` and `set_alias`
    methods. If those are used to set resolvers for interface's fields, those
    resolvers will instead be set on fields of GraphQL types implementing this
    interface, but only if those fields don't already have resolver of their own set
    by the `ObjectType`.

    # Type resolver

    Because GraphQL fields using interface as their returning type can return any
    Python value from their resolver, GraphQL interfaces require special type of
    resolver called "type resolver" to function.

    This resolver is called with the value returned by field's resolver and is
    required to return a string with a name of GraphQL type represented by Python
    value from the field:

    ```python
    def example_type_resolver(obj: Any, *_) -> str:
        if isinstance(obj, PythonReprOfUser):
            return "User"

        if isinstance(obj, PythonReprOfComment):
            return "Comment"

        raise ValueError(f"Don't know GraphQL type for '{obj}'!")
    ```

    This resolver is not required if the GraphQL field returns a value that has
    the `__typename` attribute or `dict` key with a name of the GraphQL type:

    ```python
    user_data_dict = {"__typename": "User", ...}

    # or...

    class UserRepr:
        __typename: str = "User"
    ```

    # Example

    Following code creates a GraphQL schema with a field that returns random
    result of either `User` or `Post` GraphQL type. It also supports dict with
    `__typename` key that explicitly declares its GraphQL type:

    ```python
    import random
    from dataclasses import dataclass
    from ariadne import QueryType, InterfaceType, make_executable_schema

    @dataclass
    class UserModel:
        id: str
        name: str

    @dataclass
    class PostModel:
        id: str
        message: str

    results = (
        UserModel(id=1, name="Bob"),
        UserModel(id=2, name="Alice"),
        UserModel(id=3, name="Jon"),
        PostModel(id=1, message="Hello world!"),
        PostModel(id=2, message="How's going?"),
        PostModel(id=3, message="Sure thing!"),
        {"__typename": "User", "id": 4, "name": "Polito"},
        {"__typename": "User", "id": 5, "name": "Aerith"},
        {"__typename": "Post", "id": 4, "message": "Good day!"},
        {"__typename": "Post", "id": 5, "message": "Whats up?"},
    )

    query_type = QueryType()

    @query_type.field("result")
    def resolve_random_result(*_):
        return random.choice(results)


    result_type = InterfaceType("Result")

    @result_type.type_resolver
    def resolve_result_type(obj: UserModel | PostModel | dict, *_) -> str:
        if isinstance(obj, UserModel):
            return "User"

        if isinstance(obj, PostModel):
            return "Post"

        if isinstance(obj, dict) and obj.get("__typename"):
            return obj["__typename"]

        raise ValueError(f"Don't know GraphQL type for '{obj}'!")


    schema = make_executable_schema(
        \"\"\"
        type Query {
            result: Result!
        }

        interface Result {
            id: ID!
        }

        type User implements Result {
            id: ID!
            name: String!
        }

        type Post implements Result {
            id: ID!
            message: String!
        }
        \"\"\",
        query_type,
        result_type,
    )
    ```
    """

    _resolve_type: Optional[Resolver]

    def __init__(self, name: str, type_resolver: Optional[Resolver] = None) -> None:
        """Initializes the `InterfaceType` with a `name` and optional `type_resolver`.

        Type resolver is required by `InterfaceType` to function properly, but can
        be set later using either `set_type_resolver(type_resolver)`
        setter or `type_resolver` decorator.

        # Required arguments

        `name`: a `str` with the name of GraphQL interface type in GraphQL schema to
        bind to.

        # Optional arguments

        `type_resolver`: a `Resolver` used to resolve a str with name of GraphQL type
        from it's Python representation.
        """

        super().__init__(name)
        self._resolve_type = type_resolver

    def set_type_resolver(self, type_resolver: Resolver) -> Resolver:
        """Sets function as type resolver for this interface.

        Can be used as a decorator. Also available through `type_resolver` alias:

        ```python
        interface_type = InterfaceType("MyInterface")

        @interface_type.type_resolver
        def type_resolver(obj: Any, *_) -> str:
            ...
        ```
        """
        self._resolve_type = type_resolver
        return type_resolver

    # Alias type resolver for consistent decorator API
    type_resolver = set_type_resolver

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Binds this `InterfaceType` instance to the instance of GraphQL schema.

        Sets `resolve_type` attribute on GraphQL interface. If this attribute was
        previously set, it will be replaced to new value.

        If this interface has any resolvers set, it also scans GraphQL schema for
        types implementing this interface and sets those resolvers on those types
        fields, but only if those fields don't already have other resolver set.
        """
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)
        graphql_type = cast(GraphQLInterfaceType, graphql_type)

        graphql_type.resolve_type = self._resolve_type
        self.bind_resolvers_to_graphql_type(graphql_type)

        for object_type in schema.type_map.values():
            if type_implements_interface(self.name, object_type):
                self.bind_resolvers_to_graphql_type(object_type, replace_existing=False)

    def validate_graphql_type(self, graphql_type: Optional[GraphQLNamedType]) -> None:
        """Validates that schema's GraphQL type associated with this `InterfaceType`
        is an `interface`."""
        if not graphql_type:
            raise ValueError(f"Interface {self.name} is not defined in the schema")
        if not isinstance(graphql_type, GraphQLInterfaceType):
            raise ValueError(
                f"{self.name} is defined in the schema, but it "
                f"is instance of {type(graphql_type).__name__} "
                f"(expected {GraphQLInterfaceType.__name__})"
            )



from logging import Logger, LoggerAdapter
from typing import Any, Awaitable, Optional, Type, Union

from graphql import ExecutionContext, GraphQLSchema
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import Receive, Scope, Send

from ..explorer import Explorer, ExplorerGraphiQL
from ..format_error import format_error
from ..types import (
    ContextValue,
    ErrorFormatter,
    QueryParser,
    RootValue,
    ValidationRules,
)
from .handlers import (
    GraphQLHTTPHandler,
    GraphQLWebsocketHandler,
    GraphQLWSHandler,
)


class GraphQL:
    """ASGI application implementing the GraphQL server.

    Can be used stand-alone or mounted within other ASGI application, for
    example in Starlette or FastAPI.
    """

    def __init__(
        self,
        schema: GraphQLSchema,
        *,
        context_value: Optional[ContextValue] = None,
        root_value: Optional[RootValue] = None,
        query_parser: Optional[QueryParser] = None,
        validation_rules: Optional[ValidationRules] = None,
        debug: bool = False,
        introspection: bool = True,
        explorer: Optional[Explorer] = None,
        logger: Union[None, str, Logger, LoggerAdapter] = None,
        error_formatter: ErrorFormatter = format_error,
        execution_context_class: Optional[Type[ExecutionContext]] = None,
        http_handler: Optional[GraphQLHTTPHandler] = None,
        websocket_handler: Optional[GraphQLWebsocketHandler] = None,
    ) -> None:
        """Initializes the ASGI app and it's http and websocket handlers.

        # Required arguments

        `schema`: an instance of GraphQL schema to execute queries against.

        # Optional arguments

        `context_value`: a `ContextValue` to use by this server for context.
        Defaults to `{"request": request}` dictionary where `request` is
        an instance of `starlette.requests.Request`.

        `root_value`: a `RootValue` to use by this server for root value.
        Defaults to `None`.

        `query_parser`: a `QueryParser` to use by this server. Defaults to
        `graphql.parse`.

        `validation_rules`: a `ValidationRules` list or callable returning a
        list of extra validation rules server should use to validate the
        GraphQL queries. Defaults to `None`.

        `debug`: a `bool` controlling in server should run in debug mode or
        not. Controls details included in error data returned to clients.
        Defaults to `False`.

        `introspection`: a `bool` controlling if server should allow the
        GraphQL introspection queries. If `False`, introspection queries will
        fail to pass the validation. Defaults to `True`.

        `explorer`: an instance of `Explorer` subclass to use when the server
        receives an HTTP GET request. If not set, default GraphQL explorer
        for your version of Ariadne is used.

        `logger`: a `str` with name of logger or logger instance server
        instance should use for logging errors. If not set, a logger named
        `ariadne` is used.

        `error_formatter`: an `ErrorFormatter` this server should use to format
        GraphQL errors returned to clients. If not set, default formatter
        implemented by Ariadne is used.

        `execution_context_class`: custom `ExecutionContext` type to use by
        this server to execute the GraphQL queries. Defaults to standard
        context type implemented by the `graphql`.

        `http_handler`: an instance of `GraphQLHTTPHandler` class implementing
        the HTTP requests handling logic for this server. If not set,
        an instance of `GraphQLHTTPHandler` is used.

        `websocket_handler`: an instance of `GraphQLWebsocketHandler` class
        implementing the websocket connections handling logic for this server.
        If not set, `GraphQLWSHandler` will be used, implementing older
        version of GraphQL subscriptions protocol.
        """
        if http_handler:
            self.http_handler = http_handler
        else:
            self.http_handler = GraphQLHTTPHandler()

        if websocket_handler:
            self.websocket_handler = websocket_handler
        else:
            self.websocket_handler = GraphQLWSHandler()

        if not explorer:
            explorer = ExplorerGraphiQL()

        self.http_handler.configure(
            schema,
            context_value,
            root_value,
            query_parser,
            validation_rules,
            debug,
            introspection,
            explorer,
            logger,
            error_formatter,
            execution_context_class,
        )
        self.websocket_handler.configure(
            schema,
            context_value,
            root_value,
            query_parser,
            validation_rules,
            debug,
            introspection,
            explorer,
            logger,
            error_formatter,
            execution_context_class,
            http_handler=self.http_handler,
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """An entrypoint to the ASGI application.

        Supports both HTTP and WebSocket connections.

        # Required arguments

        `scope`: The connection scope information, a dictionary that contains
        at least a type key specifying the protocol that is incoming.

        `receive`: an awaitable callable that will yield a new event dictionary
        when one is available.

        `send`: an awaitable callable taking a single event dictionary as a
        positional argument that will return once the send has been completed
        or the connection has been closed.

        Details about the arguments and their usage are described in the
        ASGI specification:

        https://asgi.readthedocs.io/en/latest/specs/main.html
        """
        if scope["type"] == "http":
            await self.http_handler.handle(scope=scope, receive=receive, send=send)
        elif scope["type"] == "websocket":
            await self.websocket_handler.handle(scope=scope, receive=receive, send=send)
        else:
            raise ValueError("Unknown scope type: %r" % (scope["type"],))

    async def handle_request(self, request: Request) -> Response:
        """Shortcut for `graphql_app.http_handler.handle_request(...)`."""
        return await self.http_handler.handle_request(request)

    async def handle_websocket(self, websocket: Any) -> Awaitable[Any]:
        """Shortcut for `graphql_app.websocket_handler.handle_websocket(...)`."""
        return await self.websocket_handler.handle_websocket(websocket)



import os
from typing import Optional, Union

from .constants import HTTP_STATUS_400_BAD_REQUEST


class HttpError(Exception):
    """Base class for HTTP errors raised inside the ASGI and WSGI servers."""

    status = ""

    def __init__(self, message: Optional[str] = None) -> None:
        """Initializes the `HttpError` with optional error message.

        # Optional arguments

        `message`: a `str` with error message to return in response body or
        `None`.
        """
        super().__init__()
        self.message = message


class HttpBadRequestError(HttpError):
    """Raised when request did not contain the data required to execute
    the GraphQL query."""

    status = HTTP_STATUS_400_BAD_REQUEST


class GraphQLFileSyntaxError(Exception):
    """Raised by `load_schema_from_path` when loaded GraphQL file has invalid syntax."""

    def __init__(self, file_path: Union[str, os.PathLike], message: str) -> None:
        """Initializes the `GraphQLFileSyntaxError` with file name and error.

        # Required arguments

        `file_path`: a `str` or `PathLike` object pointing to a file that
        failed to validate.

        `message`: a `str` with validation message.
        """
        super().__init__()

        self.message = self.format_message(file_path, message)

    def format_message(self, file_path: Union[str, os.PathLike], message: str):
        """Builds final error message from path to schema file and error message.

        Returns `str` with final error message.

        # Required arguments

        `file_path`: a `str` or `PathLike` object pointing to a file that
        failed to validate.

        `message`: a `str` with validation message.
        """
        return f"Could not load {file_path}:\n{message}"

    def __str__(self):
        """Returns error message."""
        return self.message



import enum
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from functools import reduce, singledispatch
import operator

from graphql.type import GraphQLEnumType, GraphQLNamedType, GraphQLSchema
from graphql.language.ast import (
    EnumValueNode,
    InputValueDefinitionNode,
    ObjectValueNode,
)
from graphql.pyutils.undefined import Undefined
from graphql.type.definition import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLInterfaceType,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
)

from .types import SchemaBindable


T = TypeVar("T")
ArgumentWithKeys = Tuple[str, str, GraphQLArgument, Optional[List["str"]]]
InputFieldWithKeys = Tuple[str, str, GraphQLInputField, Optional[List["str"]]]
GraphQLNamedInputType = Union[
    GraphQLScalarType, GraphQLEnumType, GraphQLInputObjectType
]


class EnumType(SchemaBindable):
    """Bindable mapping Python values to enumeration members in a GraphQL schema.

    # Example

    Given following GraphQL enum:

    ```graphql
    enum UserRole {
        MEMBER
        MODERATOR
        ADMIN
    }
    ```

    You can use `EnumType` to map it's members to Python `Enum`:

    ```python
    user_role_type = EnumType(
        "UserRole",
        {
            "MEMBER": 0,
            "MODERATOR": 1,
            "ADMIN": 2,
        }
    )
    ```

    `EnumType` also works with dictionaries:

    ```python
    user_role_type = EnumType(
        "UserRole",
        {
            "MEMBER": 0,
            "MODERATOR": 1,
            "ADMIN": 2,
        }
    )
    ```
    """

    def __init__(
        self,
        name: str,
        values: Union[Dict[str, Any], Type[enum.Enum], Type[enum.IntEnum]],
    ) -> None:
        """Initializes the `EnumType` with `name` and `values` mapping.

        # Required arguments

        `name`: a `str` with the name of GraphQL enum type in GraphQL schema to
        bind to.

        `values`: a `dict` or `enums.Enum` with values to use to represent GraphQL
        enum's in Python logic.
        """
        self.name = name
        self.values = cast(Dict[str, Any], getattr(values, "__members__", values))

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Binds this `EnumType` instance to the instance of GraphQL schema."""
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)
        graphql_type = cast(GraphQLEnumType, graphql_type)

        for key, value in self.values.items():
            if key not in graphql_type.values:
                raise ValueError(
                    "Value %s is not defined on enum %s" % (key, self.name)
                )
            graphql_type.values[key].value = value

    def bind_to_default_values(self, schema: GraphQLSchema) -> None:
        """Populates default values of input fields and args in the GraphQL schema.

        This step is required because GraphQL query executor doesn't perform a
        lookup for default values defined in schema. Instead it simply pulls the
        value from fields and arguments `default_value` attribute, which is
        `None` by default.
        """
        for _, _, arg, key_list in find_enum_values_in_schema(schema):
            type_ = resolve_null_type(arg.type)
            type_ = cast(GraphQLNamedInputType, type_)

            if (
                key_list is None
                and arg.default_value in self.values
                and type_.name == self.name
            ):
                type_ = resolve_null_type(arg.type)
                arg.default_value = self.values[arg.default_value]

            elif key_list is not None:
                enum_value = get_value_from_mapping_value(arg.default_value, key_list)
                type_ = cast(GraphQLEnumType, track_type_for_nested(arg, key_list))

                if enum_value in self.values and type_.name == self.name:
                    set_leaf_value_in_mapping(
                        arg.default_value, key_list, self.values[enum_value]
                    )

    def validate_graphql_type(self, graphql_type: Optional[GraphQLNamedType]) -> None:
        """Validates that schema's GraphQL type associated with this `EnumType`
        is an `enum`."""
        if not graphql_type:
            raise ValueError("Enum %s is not defined in the schema" % self.name)
        if not isinstance(graphql_type, GraphQLEnumType):
            raise ValueError(
                "%s is defined in the schema, but it is instance of %s (expected %s)"
                % (self.name, type(graphql_type).__name__, GraphQLEnumType.__name__)
            )


def set_default_enum_values_on_schema(schema: GraphQLSchema):
    """Sets missing Python values for GraphQL enums in schema.

    Recursively scans GraphQL schema for enums and their values. If `value`
    attribute is empty, its populated with with a string of its GraphQL name.

    This string is then used to represent enum's value in Python instead of `None`.

    # Requires arguments

    `schema`: a GraphQL schema to set enums default values in.
    """
    for type_object in schema.type_map.values():
        if isinstance(type_object, GraphQLEnumType):
            set_default_enum_values(type_object)


def set_default_enum_values(graphql_type: GraphQLEnumType):
    for key in graphql_type.values:
        if graphql_type.values[key].value is None:
            graphql_type.values[key].value = key


def validate_schema_enum_values(schema: GraphQLSchema) -> None:
    """Raises `ValueError` if GraphQL schema has input fields or arguments with
    default values that are undefined enum values.

    # Example schema with invalid field argument

    This schema fails to validate because argument `role` on field `users`
    specifies `REVIEWER` as default value and `REVIEWER` is not a member of
    the `UserRole` enum:

    ```graphql
    type Query {
        users(role: UserRole = REVIEWER): [User!]!
    }

    enum UserRole {
        MEMBER
        MODERATOR
        ADMIN
    }

    type User {
        id: ID!
    }
    ```

    # Example schema with invalid input field

    This schema fails to validate because field `role` on input `UserFilters`
    specifies `REVIEWER` as default value and `REVIEWER` is not a member of
    the `UserRole` enum:

    ```graphql
    type Query {
        users(filter: UserFilters): [User!]!
    }

    input UserFilters {
        name: String
        role: UserRole = REVIEWER
    }

    enum UserRole {
        MEMBER
        MODERATOR
        ADMIN
    }

    type User {
        id: ID!
    }
    ```
    """

    for type_name, field_name, arg, _ in find_enum_values_in_schema(schema):
        if is_invalid_enum_value(arg):
            raise ValueError(
                f"Value for type: <{arg.type}> is invalid. "
                f"Check InputField/Arguments for <{field_name}> in <{type_name}> "
                "(Undefined enum value)."
            )


def is_invalid_enum_value(field: Union[GraphQLInputField, GraphQLArgument]) -> bool:
    if field.ast_node is None:
        return False
    return field.default_value is Undefined and field.ast_node.default_value is not None


def find_enum_values_in_schema(
    schema: GraphQLSchema,
) -> Generator[Union[ArgumentWithKeys, InputFieldWithKeys], None, None]:
    for name, type_ in schema.type_map.items():
        result = enum_values_in_types(type_, name)
        if result is not None:
            yield from result


@singledispatch
def enum_values_in_types(
    type_: GraphQLNamedType,  # pylint: disable=unused-argument
    name: str,  # pylint: disable=unused-argument
) -> Optional[Generator[Union[ArgumentWithKeys, InputFieldWithKeys], None, None]]:
    pass


@enum_values_in_types.register(GraphQLObjectType)
@enum_values_in_types.register(GraphQLInterfaceType)
def enum_values_in_object_type(
    type_: Union[GraphQLObjectType, GraphQLInterfaceType],
    field_name: str,
) -> Generator[ArgumentWithKeys, None, None]:
    for field in type_.fields.values():
        yield from enum_values_in_field_args(field_name, field)


@enum_values_in_types.register(GraphQLInputObjectType)
def enum_values_in_input_type(
    type_: GraphQLInputObjectType,
    field_name,
) -> Generator[InputFieldWithKeys, None, None]:
    yield from _get_field_with_keys(field_name, type_.fields.items())


def enum_values_in_field_args(
    field_name: str,
    field: GraphQLField,
) -> Generator[ArgumentWithKeys, None, None]:
    args = [
        (name, arg)
        for name, arg in field.args.items()
        if isinstance(
            arg.type, (GraphQLInputObjectType, GraphQLEnumType, GraphQLNonNull)
        )
    ]

    yield from _get_field_with_keys(field_name, args)


def _get_field_with_keys(field_name, fields):
    for input_name, field in fields:
        resolved_type = resolve_null_type(field.type)
        if isinstance(resolved_type, GraphQLEnumType):
            yield field_name, input_name, field, None

        if isinstance(resolved_type, GraphQLInputObjectType):
            if field.ast_node is not None and field.ast_node.default_value is not None:
                routes = get_enum_keys_from_ast(field.ast_node)
                for route in routes:
                    yield field_name, input_name, field, route


def get_enum_keys_from_ast(ast_node: InputValueDefinitionNode) -> List[List["str"]]:
    routes = []
    object_node = cast(ObjectValueNode, ast_node.default_value)
    nodes = [([field.name.value], field) for field in object_node.fields]

    while nodes:
        key_list, field = nodes.pop()
        if isinstance(field.value, EnumValueNode):
            routes.append(key_list)

        if isinstance(field.value, ObjectValueNode):
            for new_field in field.value.fields:
                new_route = key_list[:]
                new_route.append(new_field.name.value)
                nodes.append((new_route, new_field))

    return routes


def get_value_from_mapping_value(mapping: Mapping[T, Any], key_list: List[T]) -> Any:
    return reduce(operator.getitem, key_list, mapping)


def set_leaf_value_in_mapping(
    mapping: Mapping[T, Any], key_list: List[T], value: Any
) -> None:
    get_value_from_mapping_value(mapping, key_list[:-1])[key_list[-1]] = value


def track_type_for_nested(
    arg: Union[GraphQLArgument, GraphQLInputField], key_list: List[str]
) -> GraphQLInputType:
    type_ = resolve_null_type(arg.type)

    for elem in key_list:
        if isinstance(type_, GraphQLInputObjectType):
            type_ = type_.fields[elem].type
    return type_


def resolve_null_type(type_: GraphQLInputType) -> GraphQLInputType:
    return type_.of_type if isinstance(type_, GraphQLNonNull) else type_



from typing import Optional, cast

from graphql.type import (
    GraphQLNamedType,
    GraphQLScalarLiteralParser,
    GraphQLScalarSerializer,
    GraphQLScalarType,
    GraphQLScalarValueParser,
    GraphQLSchema,
)

from .types import SchemaBindable


class ScalarType(SchemaBindable):
    """Bindable populating scalars in a GraphQL schema with Python logic.

    GraphQL scalars implement default serialization and deserialization logic.
    This class is only useful when custom logic is needed, most commonly
    when Python representation of scalar's value is not JSON-serializable by
    default.

    This logic can be customized for three steps:

    # Serialization

    Serialization step converts Python representation of scalar's value to a
    JSON serializable format.

    Serializer function takes single argument and returns a single,
    JSON serializable value:

    ```python
    def serialize_date(value: date) -> str:
        # Serialize dates as "YYYY-MM-DD" string
        return date.strftime("%Y-%m-%d")
    ```

    # Value parsing

    Value parsing step converts value from deserialized JSON
    to Python representation.

    Value parser function takes single argument and returns a single value:

    ```python
    def parse_date_str(value: str) -> date:
        try:
            # Parse "YYYY-MM-DD" string into date
            return datetime.strptime(value, "%Y-%m-%d").date()
        except (ValueError, TypeError):
            raise ValueError(
                f'"{value}" is not a date string in YYYY-MM-DD format.'
            )
    ```

    # Literal parsing

    Literal parsing step converts value from GraphQL abstract syntax tree (AST)
    to Python representation.

    Literal parser function takes two arguments, an AST node and a dict with
    query's variables and returns Python value:

    ```python
    def parse_date_literal(
        value: str, variable_values: dict[str, Any] = None
    ) -> date:
        if not isinstance(ast, StringValueNode):
            raise ValueError()

        try:
            # Parse "YYYY-MM-DD" string into date
            return datetime.strptime(ast.value, "%Y-%m-%d").date()
        except (ValueError, TypeError):
            raise ValueError(
                f'"{value}" is not a date string in YYYY-MM-DD format.'
            )
    ```

    When scalar has custom value parser set, but not the literal parser, the
    GraphQL query executor will use default literal parser, and then call the
    value parser with it's return value. This mechanism makes custom literal
    parser unnecessary for majority of scalar implementations.

    Scalar literals are always parsed twice: on query validation and during
    query execution.

    # Example datetime scalar

    Following code defines a datetime scalar which converts Python datetime
    object to and from a string. Note that it without custom literal scalar:

    ```python
    from datetime import datetime

    from ariadne import QueryType, ScalarType, make_executable_schema

    scalar_type = ScalarType("DateTime")

    @scalar_type.serializer
    def serialize_value(val: datetime) -> str:
        return datetime.strftime(val, "%Y-%m-%d %H:%M:%S")


    @scalar_type.value_parser
    def parse_value(val) -> datetime:
        if not isinstance(val, str):
            raise ValueError(
                f"'{val}' is not a valid JSON representation "
            )

        return datetime.strptime(val, "%Y-%m-%d %H:%M:%S")


    query_type = QueryType()

    @query_type.field("now")
    def resolve_now(*_):
        return datetime.now()


    @query_type.field("diff")
    def resolve_diff(*_, value):
        delta = datetime.now() - value
        return int(delta.total_seconds())


    schema = make_executable_schema(
        \"\"\"
        scalar DateTime

        type Query {
            now: DateTime!
            diff(value: DateTime): Int!
        }
        \"\"\",
        scalar_type,
        query_type,
    )
    ```

    # Example generic scalar

    Generic scalar is a pass-through scalar that doesn't perform any value
    conversion. Most common use case for those is for GraphQL fields that
    return unstructured JSON to the client. To create a scalar like this,
    you can simply include  `scalar Generic` in your GraphQL schema:

    ```python
    from ariadne import QueryType, make_executable_schema

    query_type = QueryType()

    @query_type.field("rawJSON")
    def resolve_raw_json(*_):
        # Note: this value needs to be JSON serializable
        return {
            "map": {
                "0": "Hello!",
                "1": "World!",
            },
            "list": [
                2,
                1,
                3,
                7,
            ],
        }


    schema = make_executable_schema(
        \"\"\"
        scalar Generic

        type Query {
            rawJSON: Generic!
        }
        \"\"\",
        query_type,
    )
    ```
    """

    _serialize: Optional[GraphQLScalarSerializer]
    _parse_value: Optional[GraphQLScalarValueParser]
    _parse_literal: Optional[GraphQLScalarLiteralParser]

    def __init__(
        self,
        name: str,
        *,
        serializer: Optional[GraphQLScalarSerializer] = None,
        value_parser: Optional[GraphQLScalarValueParser] = None,
        literal_parser: Optional[GraphQLScalarLiteralParser] = None,
    ) -> None:
        """Initializes the `ScalarType` with a `name`.

        # Required arguments

        `name`: a `str` with the name of GraphQL scalar in GraphQL schema to
        bind to.

        # Optional arguments

        `serializer`: a function called to convert Python representation of
        scalar's value to JSON serializable format.

        `value_parser`: a function called to convert a JSON deserialized value
        from query's "variables" JSON into scalar's Python representation.

        `literal_parser`: a function called to convert an AST value
        from parsed query into scalar's Python representation.
        """
        self.name = name
        self._serialize = serializer
        self._parse_value = value_parser
        self._parse_literal = literal_parser

    def set_serializer(self, f: GraphQLScalarSerializer) -> GraphQLScalarSerializer:
        """Sets function as serializer for this scalar.

        Can be used as a decorator. Also available through `serializer` alias:

        ```python
        date_scalar = ScalarType("Date")

        @date_scalar.serializer
        def serialize_date(value: date) -> str:
            # Serialize dates as "YYYY-MM-DD" string
            return date.strftime("%Y-%m-%d")
        ```
        """
        self._serialize = f
        return f

    def set_value_parser(self, f: GraphQLScalarValueParser) -> GraphQLScalarValueParser:
        """Sets function as value parser for this scalar.

        Can be used as a decorator. Also available through `value_parser` alias:

        ```python
        date_scalar = ScalarType("Date")

        @date_scalar.value_parser
        def parse_date_str(value: str) -> date:
            try:
                # Parse "YYYY-MM-DD" string into date
                return datetime.strptime(value, "%Y-%m-%d").date()
            except (ValueError, TypeError):
                raise ValueError(
                    f'"{value}" is not a date string in YYYY-MM-DD format.'
                )
        ```
        """
        self._parse_value = f
        return f

    def set_literal_parser(
        self, f: GraphQLScalarLiteralParser
    ) -> GraphQLScalarLiteralParser:
        """Sets function as literal parser for this scalar.

        Can be used as a decorator. Also available through `literal_parser` alias:

        ```python
        date_scalar = ScalarType("Date")

        @date_scalar.literal_parser
        def parse_date_literal(
            value: str, variable_values: Optional[dict[str, Any]] = None
        ) -> date:
            if not isinstance(ast, StringValueNode):
                raise ValueError()

            try:
                # Parse "YYYY-MM-DD" string into date
                return datetime.strptime(ast.value, "%Y-%m-%d").date()
            except (ValueError, TypeError):
                raise ValueError(
                    f'"{value}" is not a date string in YYYY-MM-DD format.'
                )
        ```
        """
        self._parse_literal = f
        return f

    # Alias above setters for consistent decorator API
    serializer = set_serializer
    value_parser = set_value_parser
    literal_parser = set_literal_parser

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Binds this `ScalarType` instance to the instance of GraphQL schema.

        If it has serializer or parser functions set, it assigns those to GraphQL
        scalar's attributes. If scalar's attribute already has other function
        set, this function is replaced with the new one.
        """
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)
        graphql_type = cast(GraphQLScalarType, graphql_type)

        if self._serialize:
            # See mypy bug https://github.com/python/mypy/issues/2427
            graphql_type.serialize = self._serialize  # type: ignore
        if self._parse_value:
            graphql_type.parse_value = self._parse_value  # type: ignore
        if self._parse_literal:
            graphql_type.parse_literal = self._parse_literal  # type: ignore

    def validate_graphql_type(self, graphql_type: Optional[GraphQLNamedType]) -> None:
        """Validates that schema's GraphQL type associated with this `ScalarType`
        is a `scalar`."""
        if not graphql_type:
            raise ValueError("Scalar %s is not defined in the schema" % self.name)
        if not isinstance(graphql_type, GraphQLScalarType):
            raise ValueError(
                "%s is defined in the schema, but it is instance of %s (expected %s)"
                % (self.name, type(graphql_type).__name__, GraphQLScalarType.__name__)
            )



from inspect import isawaitable
from dataclasses import dataclass
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Collection,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)
from typing_extensions import Protocol, runtime_checkable

from graphql import (
    DocumentNode,
    ExecutionResult,
    GraphQLError,
    GraphQLResolveInfo,
    GraphQLSchema,
)
from graphql.validation.rules import ASTValidationRule

from starlette.websockets import WebSocket

__all__ = [
    "Resolver",
    "GraphQLResult",
    "SubscriptionResult",
    "Subscriber",
    "ErrorFormatter",
    "ContextValue",
    "RootValue",
    "QueryParser",
    "ValidationRules",
    "ExtensionList",
    "Extensions",
    "Middleware",
    "MiddlewareList",
    "Middlewares",
    "Operation",
    "OnConnect",
    "OnDisconnect",
    "OnOperation",
    "OnComplete",
    "WebSocketConnectionError",
    "Extension",
    "ExtensionSync",
    "SchemaBindable",
]


# Note: this should be [Any, GraphQLResolveInfo, **kwargs],
# but this is not achieveable with python types yet:
# https://github.com/mirumee/ariadne/pull/79
"""Type for resolver functions.

Due to limitations in Mypy this type is unspecific. More accurate type is:

```python
Resolver = Callable[[Any, GraphQLResolveInfo, KwArg[Any]], Any]
```

Each resolver is called with two positional arguments and any number of keyword 
arguments:

`Any`: a Python representation of GraphQL type to resolve value from.

`GraphQLResolveInfo`: an instance of GraphQL object holding data about currently 
executed GraphQL field, and `context` attribute.

`KwArg[Any]`: if currently executed GraphQL field has any arguments, their 
values will be passed as keyword arguments.
"""
Resolver = Callable[..., Any]

"""Result type for `graphql` and `graphql_sync` functions.

It's a tuple of two elements:

`bool`: `True` when query was executed successfully (without any errors), `False` otherwise.

`dict`: JSON-serializable query result.
"""
GraphQLResult = Tuple[bool, dict]

"""Result type for `subscribe` function.

It's a tuple of two elements:

`bool`: `True` when query was executed successfully (without any errors), `False` otherwise.

`dict or generator`: JSON-serializable query result or asynchronous generator with 
subscription's results. Depends if query was success or not.
"""
SubscriptionResult = Tuple[
    bool, Union[List[dict], AsyncGenerator[ExecutionResult, None]]
]

"""Type for subscription source functions.

Due to limitations in Mypy this type is unspecific. More accurate type is:

```python
Subscriber = Callable[[Any, GraphQLResolveInfo, KwArg[Any]], AsyncGenerator]
```

Receives same arguments as `Resolver`.
"""
Subscriber = Callable[..., AsyncGenerator]

"""Type for custom error formatters.

Error formatter is a function called by Ariadne to convert `GraphQLError` 
instance into JSON-serializable dict to return to client.

It receives two arguments:

`GraphQLError`: an error to serialize.

`bool`: a `debug` flag, which tells the formatter if it should (`True`) or 
shouldn't (`False`) include debugging information.
"""
ErrorFormatter = Callable[[GraphQLError, bool], dict]

"""Type for `context_value` option of GraphQL servers.

Context value is accessible in GraphQL resolvers as `context` attribute of 
second argument:

```python
def resolve_example(_, info: GraphQLResolveInfo):
    info.context  # Do something with context
```

# Default context

Default context value passed to resolvers by Ariadne is a dictionary with single 
key, `request`, which contains the HTTP framework specific representation of HTTP 
request:

```python
def resolve_example(_, info: GraphQLResolveInfo):
    request = info.context["request"]  # Get request from context
```

# Dynamic context value

If context value is a callable, it will be evaluated at the beginning of query 
execution.

It's called with two arguments:

`request`: an representation of HTTP request specific to the framework used.

`data`: an _unvalidated_ JSON which may be a valid GraphQL request payload.

Callable can return any value which will then be passed to resolvers. Some 
implementations (like `ariadne.asgi.GraphQL`) support this callable being 
asynchronous.

Here's an example callable that retrieves application settings from database, 
then uses them together with request to retrieve authenticated user. Finally 
it exposes all three of those to GraphQL resolvers:

```python
from ariadne.asgi import GraphQL
from starlette.requests import Request

from .auth import get_authenticated_user
from .settings import get_db_settings
from .schema import schema

async def get_context_value(request: Request, _):
    settings = await get_db_settings()
    user = await get_authenticated_user(request, settings)

    return {
        "request"; request,
        "settings": settings,
        "user": user,
    }

graphql_app = GraphQL(
    schema,
    context_value=get_context_value,
)
```
"""
ContextValue = Union[
    Any,
    Callable[[Any], Any],  # TODO: remove in 0.20
    Callable[[Any, dict], Any],
]

"""Type for `root_value` option of GraphQL servers.

"Root value" is a value passed to root resolvers (resolvers set on `Query`, 
`Mutation` and `Subscription` fields) as first argument.

# Default root value

Ariadne doesn't define root value by default. First argument of root resolvers 
is `None`.

# Dynamic root value

If root value is a callable, it will be evaluated at the beginning of query 
handling. It's called with four arguments:

`context_value`: a context value specific to this GraphQL server.

`operation_name`: a `str` with name of operation to execute (or `None`).

`variables`: a `dict` with variables to pass to query's resolvers (or `None`).

`document`: a `DocumentNode` with parsed GraphQL query.

Callable can return any value which then will be passed to root resolvers. 
Some implementations (like `ariadne.asgi.GraphQL`) support this callable being 
asynchronous.
"""
RootValue = Union[
    Any,
    Callable[[Optional[Any], DocumentNode], Any],  # TODO: remove in 0.20
    Callable[[Optional[Any], Optional[str], Optional[dict], DocumentNode], Any],
]

"""Type of `query_parser` option of GraphQL servers.

Enables customization of server's GraphQL parsing logic. If not set or `None`, 
default parser is used instead.

# Default query parser

Default query parser used by Ariadne is `parse` function from the `graphql` 
package.

# Custom parser

Custom parser is a function or callable accepting two arguments:

`context_value`: a context value specific to this GraphQL server.

`data`: a `dict` with validated GraphQL request data (contains `query` string, 
optionally also has `operationName` string or `variables` dictionary).

Parser is required to return `DocumentNode` with parsed query or raise 
`GraphQLError` when query is data invalid .

Asynchronous parsers are __not__ supported.

# Example parser

Below code defines custom parser that discards query string altogether, using 
predefined queries instead:

```python
from graphql import GraphQLError, parse


class AllowedQueriesParser:
    def __init__(self):
        self._queries = {}

    def __call__(self, _, data):
        operation_name = data.get("operationName")
        if not operation_name:
            raise GraphQLError(
                "Explicit 'operationName' is required by the server."
            )
        if operation_name not in self._queries:
            raise GraphQLError(
                f"Operation 'operation_name' is not supported by the server."
            )

        return self._queries[operation_name]

    def add_query(self, operation_name: str, query: str):
        self._queries[operation_name] = parse(query)


allowed_queries_parser = AllowedQueriesParser()

allowed_queries_parser.add_query(
    "GetUsers",
    \"\"\"
    query GetUsers {
        users {
            id
            name
        }
    }
    \"\"\"
)
```
"""
QueryParser = Callable[[ContextValue, Dict[str, Any]], DocumentNode]

"""Type of `validation_rules` option of GraphQL servers.

Enables extending of server's query validation logic.

Is either a callable that should return list of validation rules to use for 
GraphQL request, or list of validation rules.

Callable is evaluated with three arguments:

`Optional[Any]`: a context value for this request, or `None`.

`DocumentNode`: a `document` with parsed GraphQL query.

`dict`: a GraphQL request's data.
"""
ValidationRules = Union[
    Collection[Type[ASTValidationRule]],
    Callable[
        [Optional[Any], DocumentNode, dict],
        Optional[Collection[Type[ASTValidationRule]]],
    ],
]

"""List of extensions to use during GraphQL query execution."""
ExtensionList = Optional[List[Union[Type["Extension"], Callable[[], "Extension"]]]]

"""Type of `extensions` option of GraphQL servers.

It's either a list of extensions (see `ExtensionList`), or callable that 
returns this list.

Callable is evaluated with two arguments:

`Any`: the HTTP framework specific representation of HTTP request.

`Optional[ContextValue]`: a context value for this request, or `None`.
"""
Extensions = Union[
    Callable[[Any, Optional[ContextValue]], ExtensionList], ExtensionList
]

# Unspecific Middleware type in line what graphql-core expects.
# Could be made more specific in future versions but currently MyPY doesn't
# handle mixing __positional __args with **kwargs that we need.
# Callable[[Resolver, Any, GraphQLResolveInfo, KwArg(Any)], Any]
"""GraphQL middleware.

Due to limitations in Mypy this type is unspecific. More accurate type is:

```python
Middleware = Callable[[Resolver, Any, GraphQLResolveInfo, KwArg[Any]], Any]
```

Each middleware is called with three positional arguments and any number of 
keyword arguments:

`Resolver`: executed field's resolver or next GraphQL middleware.

`Any`: a Python representation of GraphQL type to resolve value from.

`GraphQLResolveInfo`: an instance of GraphQL object holding data about currently 
executed GraphQL field, and `context` attribute.

`KwArg[Any]`: if currently executed GraphQL field has any arguments, their 
values will be passed as keyword arguments.
"""
Middleware = Callable[..., Any]

"""List of middlewares to use during GraphQL query execution."""
MiddlewareList = Optional[Sequence[Middleware]]

"""Type of `middleware` option of GraphQL servers.

It's either a list of middleware (see `MiddlewareList`), or callable that 
returns this list.

Callable is evaluated with two arguments:

`Any`: the HTTP framework specific representation of HTTP request.

`Optional[ContextValue]`: a context value for this request, or `None`.
"""
Middlewares = Union[
    Callable[[Any, Optional[ContextValue]], MiddlewareList], MiddlewareList
]


@dataclass
class Operation:
    """Dataclass representing single active GraphQL operation."""

    id: str
    name: Optional[str]
    generator: AsyncGenerator


"""Type of `on_connect` option of GraphQL websocket servers.

Callback function evaluated when new websocket connection was established. 
Usually used to update connection's `scope` with contents of initial message.

Called with two arguments:

`WebSocket`: the HTTP framework specific representation of websocket connection.

`Any`: a data sent in WebSocket message.
"""
OnConnect = Callable[[WebSocket, Any], Any]

"""Type of `on_disconnect` option of GraphQL websocket servers.

Callback function evaluated when websocket connection is closed.

Called with one arguments:

`WebSocket`: the HTTP framework specific representation of websocket connection.
"""
OnDisconnect = Callable[[WebSocket], Any]

"""Type of `on_operation` option of GraphQL websocket servers.

Callback function evaluated when individual GraphQL subscription is initiated.

Called with two arguments:

`WebSocket`: the HTTP framework specific representation of websocket connection.

`Operation`: an object with initiated subscription's data.
"""
OnOperation = Callable[[WebSocket, Operation], Any]

"""Type of `on_complete` option of GraphQL websocket servers.

Callback function evaluated when individual GraphQL subscription is completed.

Called with two arguments:

`WebSocket`: the HTTP framework specific representation of websocket connection.

`Operation`: an object with closed subscription's data.
"""
OnComplete = Callable[[WebSocket, Operation], Any]


class WebSocketConnectionError(Exception):
    """Special error class enabling custom error reporting for on_connect"""

    def __init__(self, payload: Optional[Union[dict, str]] = None) -> None:
        if isinstance(payload, dict):
            self.payload = payload
        elif payload:
            self.payload = {"message": str(payload)}
        else:
            self.payload = {"message": "Unexpected error has occurred."}


@runtime_checkable
class Extension(Protocol):
    """Base class for async extensions.

    Subclasses of this this class should override default methods to run
    custom logic during Query execution.
    """

    def request_started(self, context: ContextValue) -> None:
        """Extension hook executed at request's start."""

    def request_finished(self, context: ContextValue) -> None:
        """Extension hook executed at request's end."""

    async def resolve(
        self, next_: Resolver, obj: Any, info: GraphQLResolveInfo, **kwargs
    ) -> Any:
        """Async extension hook wrapping field's value resolution.

        # Arguments

        `next_`: a `resolver` or next extension's `resolve` method.

        `obj`: a Python data structure to resolve value from.

        `info`: a `GraphQLResolveInfo` instance for executed resolver.

        `**kwargs`: extra arguments from GraphQL to pass to resolver.
        """
        result = next_(obj, info, **kwargs)
        if isawaitable(result):
            result = await result
        return result

    def has_errors(self, errors: List[GraphQLError], context: ContextValue) -> None:
        """Extension hook executed when GraphQL encountered errors."""

    def format(self, context: ContextValue) -> Optional[dict]:
        """Extension hook executed to retrieve extra data to include in result's
        `extensions` data."""


class ExtensionSync(Extension):
    """Base class for sync extensions, extends `Extension`.

    Subclasses of this this class should override default methods to run
    custom logic during Query execution.
    """

    def resolve(  # pylint: disable=invalid-overridden-method
        self, next_: Resolver, obj: Any, info: GraphQLResolveInfo, **kwargs
    ) -> Any:
        """Sync extension hook wrapping field's value resolution.

        # Arguments

        `next_`: a `resolver` or next extension's `resolve` method.

        `obj`: a Python data structure to resolve value from.

        `info`: a `GraphQLResolveInfo` instance for executed resolver.

        `**kwargs`: extra arguments from GraphQL to pass to resolver.
        """
        return next_(obj, info, **kwargs)


@runtime_checkable
class SchemaBindable(Protocol):
    """Base class for bindable types.

    Subclasses should extend the `bind_to_schema` method with custom logic for
    populating an instance of GraphQL schema with Python logic and values.

    # Example

    Example `InputType` bindable that sets Python names for fields of GraphQL input:

    ```python
    from ariadne import SchemaBindable
    from graphql import GraphQLInputType

    class InputType(SchemaBindable):
        _name: str
        _fields: dict[str, str]

        def __init__(self, name: str, fields: dict[str, str] | None):
            self._name = name
            self._fields = fields or {}

        def set_field_out_name(self, field: str, out_name: str):
            self._fields[field] = out_name

        def bind_to_schema(self, schema: GraphQLSchema) -> None:
            graphql_type = schema.get_type(self._name)
            if not graphql_type:
                raise ValueError(
                    "Type %s is not defined in the schema" % self.name
                )
            if not isinstance(graphql_type, GraphQLInputType):
                raise ValueError(
                    "%s is defined in the schema, but it is instance of %s (expected %s)"
                    % (self.name, type(graphql_type).__name__, GraphQLInputType.__name__)
                )

            for field, out_name in self._fields.items():
                schema_field = graphql_type.fields.get(field)
                if not schema_field:
                    raise ValueError(
                        "Type %s does not define the %s field" % (self.name, field)
                    )

                schema_field.out_name = out_name
    ```

    Usage:

    ```python
    from ariadne import QueryType, make_executable_schema

    from .input_type import InputType
    from .users.models import User

    input_type = InputType(
        "UserInput",
        {
            "fullName": "full_name",
            "yearOfBirth": "year_of_birth",
        }
    )

    query_type = QueryType()

    @query_type.field("countUsers")
    def resolve_count_users(*_, input):
        qs = User.objects

        if input:
            if input["full_name"]:
                qs = qs.filter(full_name__ilike=input["full_name"])
            if input["year_of_birth"]:
                qs = qs.filter(dob__year=input["year_of_birth"])

        return qs.count()


    schema = make_executable_schema(
        \"\"\"
        type Query {
            countUsers(input: UserInput!): Int!
        }

        input UserInput {
            fullName: String
            yearOfBirth: Int
        }
        \"\"\",
        query_type,
        input_type,
    )
    ```
    """

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Binds this `SchemaBindable` instance to the instance of GraphQL schema."""



from .apollo import ExplorerApollo
from .default_query import escape_default_query
from .explorer import Explorer, ExplorerHttp405
from .graphiql import ExplorerGraphiQL
from .playground import ExplorerPlayground
from .template import render_template

__all__ = [
    "Explorer",
    "ExplorerApollo",
    "ExplorerGraphiQL",
    "ExplorerHttp405",
    "ExplorerPlayground",
    "escape_default_query",
    "render_template",
]



from .base import GraphQLHandler, GraphQLHttpHandlerBase, GraphQLWebsocketHandler
from .http import GraphQLHTTPHandler
from .graphql_transport_ws import GraphQLTransportWSHandler
from .graphql_ws import GraphQLWSHandler


__all__ = [
    "GraphQLHandler",
    "GraphQLHTTPHandler",
    "GraphQLHttpHandlerBase",
    "GraphQLTransportWSHandler",
    "GraphQLWSHandler",
    "GraphQLWebsocketHandler",
]



from .default_query import escape_default_query
from .explorer import Explorer
from .template import read_template, render_template

GRAPHIQL_HTML = read_template("graphiql.html")

DEFAULT_QUERY = """#
# GraphiQL is an in -browser tool for writing, validating, and
# testing GraphQL queries.
#
# Type queries into this side of the screen, and you will see intelligent
# typeaheads aware of the current GraphQL type schema and live syntax and
# validation errors highlighted within the text.
#
# GraphQL queries typically start with a "{" character. Lines that start
# with a # are ignored.
#
# An example GraphQL query might look like:
#
#     {
#       field(arg: "value") {
#         subField
#
#       }
#
#     }
#
# Keyboard shortcuts:
#
#   Prettify query: Shift - Ctrl - P(or press the prettify button)
#
#  Merge fragments: Shift - Ctrl - M(or press the merge button)
#
#        Run Query: Ctrl - Enter(or press the play button)
#
#    Auto Complete: Ctrl - Space(or just start typing)
#"""


class ExplorerGraphiQL(Explorer):
    def __init__(
        self,
        title: str = "Ariadne GraphQL",
        explorer_plugin: bool = False,
        default_query: str = DEFAULT_QUERY,
    ) -> None:
        self.parsed_html = render_template(
            GRAPHIQL_HTML,
            {
                "title": title,
                "enable_explorer_plugin": explorer_plugin,
                "default_query": escape_default_query(default_query),
            },
        )

    def html(self, _):
        return self.parsed_html



"""Tiny templating engine for explorers HTML.

Supports:

{{ varname }} - renders html.escaped varname from template vars
{% if var1 ... %}content{% endif %} - renders block contents if all args are true
{% ifnot var1 ... %}content{% endif %} - renders block contents if all args aren't true
{% else %} - used inside if/ifnot blocks to render alternative contents
{% json varname %} - renders json string with varname contents
"""

import html
from enum import IntEnum
from os import path
from typing import List, Optional, Tuple


TEMPLATE_DIR = path.join(path.dirname(path.abspath(__file__)), "templates")


def read_template(template: str) -> str:
    template_path = path.join(TEMPLATE_DIR, template)
    with open(template_path, "r", encoding="utf-8") as fp:
        return fp.read()


class Token(IntEnum):
    STR = 0
    VAR = 1
    RAW = 2
    IF = 3
    IF_NOT = 4
    ELSE = 5
    ENDIF = 6


TokenBlock = Tuple[Token, Optional[str]]


def render_template(template: str, template_vars: Optional[dict] = None) -> str:
    document = parse_template(template)
    return document.render(template_vars or {})


def parse_template(template: str):
    tokens = tokenize_template(template)
    return build_template_ast(tokens)


def tokenize_template(template: str) -> List[TokenBlock]:
    tokens: List[TokenBlock] = []
    cursor = 0
    limit = len(template)

    while cursor <= limit:
        tokens_positions = [
            pos
            for pos in [template.find("{{", cursor), template.find("{%", cursor)]
            if pos >= 0
        ]
        if tokens_positions:
            new_cursor = min(tokens_positions)
            if new_cursor > cursor:
                tokens.append((Token.STR, template[cursor:new_cursor]))
                cursor = new_cursor
        else:
            # Eject from tokenizer by appending rest of template as string
            tokens.append((Token.STR, template[cursor:limit]))
            break

        if template[cursor : cursor + 2] == "{{":
            token, cursor = tokenize_var(template, cursor)
            tokens.append(token)
        elif template[cursor : cursor + 2] == "{%":
            token, cursor = tokenize_block(template, cursor)
            tokens.append(token)

    return tokens


def tokenize_var(template: str, cursor: int) -> Tuple[TokenBlock, int]:
    end = template.find("}}", cursor)
    if end == -1:
        raise ValueError(
            f"Unclosed variable tag at {cursor}: '{template[cursor:cursor+20]}...'"
        )

    var_name = template[cursor + 2 : end].strip()
    if not var_name:
        raise ValueError(
            f"Empty variable tag at {cursor}: '{template[cursor:cursor+20]}...'"
        )

    return (Token.VAR, var_name), end + 2


def tokenize_block(template: str, cursor: int) -> Tuple[TokenBlock, int]:
    token: TokenBlock

    end = template.find("%}", cursor)
    if end == -1:
        raise ValueError(
            f"Unclosed block tag at {cursor}: '{template[cursor:cursor+20]}...'"
        )

    block_content = template[cursor + 2 : end].strip()
    if not block_content:
        raise ValueError(
            f"Empty block tag at {cursor}: '{template[cursor:cursor+20]}...'"
        )

    block_words = [word.strip() for word in block_content.split(" ")]
    block_type, block_args = block_words[0], block_words[1:]
    args = " ".join(block_args)

    if block_type.lower() == "if":
        token = (Token.IF, args)
        if not args:
            raise ValueError(
                f"'if' block without arguments at {cursor}: "
                f"'{template[cursor:cursor+20]}...'"
            )

    elif block_type.lower() == "ifnot":
        token = (Token.IF_NOT, args)
        if not args:
            raise ValueError(
                f"'ifnot' block without arguments at {cursor}: "
                f"'{template[cursor:cursor+20]}...'"
            )

    elif block_type.lower() == "else":
        token = (Token.ELSE, None)
        if args:
            raise ValueError(
                f"'else' block with arguments at {cursor}: "
                f"'{template[cursor:cursor+20]}...'"
            )

    elif block_type.lower() == "endif":
        token = (Token.ENDIF, None)
        if args:
            raise ValueError(
                f"'endif' block with arguments at {cursor}: "
                f"'{template[cursor:cursor+20]}...'"
            )

    elif block_type.lower() == "raw":
        token = (Token.RAW, args)
        if not args:
            raise ValueError(
                f"'raw' block without arguments at {cursor}: "
                f"'{template[cursor:cursor+20]}...'"
            )

    else:
        raise ValueError(
            f"Unknown block at {cursor}: '{template[cursor:cursor+20]}...'"
        )

    return token, end + 2


def build_template_ast(tokens: List[TokenBlock]) -> "TemplateDocument":
    nodes = ast_to_nodes(tokens)
    return TemplateDocument(nodes)


def ast_to_nodes(tokens: List[TokenBlock]) -> List["TemplateNode"]:
    nodes: List[TemplateNode] = []
    i = 0
    limit = len(tokens)
    while i < limit:
        token_type, token_args = tokens[i]
        if token_type == Token.STR and token_args:
            nodes.append(TemplateText(token_args))
            i += 1
            continue

        if token_type == Token.VAR and token_args:
            nodes.append(TemplateVariable(token_args))
            i += 1
            continue

        if token_type == Token.RAW and token_args:
            nodes.append(TemplateVariable(token_args, escape=False))
            i += 1
            continue

        if token_type in (Token.IF, Token.IF_NOT) and token_args:
            if i + 1 == limit:
                raise ValueError("Unclosed 'if' block found.")

            if_block_args = token_args.split(" ")
            nesting = 0
            children: List[TokenBlock] = []
            if_not = token_type == Token.IF_NOT
            has_else = False
            for child in tokens[i + 1 :]:
                i += 1
                child_type = child[0]

                if child_type == Token.ENDIF:
                    if nesting == 0:
                        i += 1
                        break

                    nesting -= 1

                elif child_type == Token.ELSE:
                    if nesting == 0:
                        if has_else:
                            raise ValueError("Multiple 'else' clauses found.")

                        nodes.append(
                            TemplateIfBlock(
                                if_block_args,
                                ast_to_nodes(children),
                                if_not,
                            )
                        )
                        children = []
                        if_not = not if_not
                        has_else = True
                        continue

                elif child_type in (Token.IF, Token.IF_NOT):
                    nesting += 1

                children.append(child)
            else:
                raise ValueError("Unclosed 'if' block found.")

            nodes.append(
                TemplateIfBlock(
                    if_block_args,
                    ast_to_nodes(children),
                    if_not,
                )
            )
            continue

        if token_type == Token.ENDIF:
            raise ValueError("Extra 'endif' block found.")

    return nodes


class TemplateNode:
    def render(self, template_vars) -> str:
        raise NotImplementedError(
            "Subclasses of TemplateNode should define 'render' method."
        )


class TemplateDocument(TemplateNode):
    def __init__(self, nodes) -> None:
        self.nodes = nodes

    def render(self, template_vars) -> str:
        return "".join([node.render(template_vars) for node in self.nodes])


class TemplateText(TemplateNode):
    def __init__(self, value: str) -> None:
        self.value = value

    def render(self, _) -> str:
        return self.value


class TemplateIfBlock(TemplateNode):
    def __init__(self, args: List[str], nodes, if_not: bool = False) -> None:
        self.args = args
        self.nodes = nodes
        self.if_not = if_not

    def render(self, template_vars) -> str:
        args_true = all(template_vars.get(arg) for arg in self.args)
        block_true = not args_true if self.if_not else args_true
        if not block_true:
            return ""

        return "".join([node.render(template_vars) for node in self.nodes])


class TemplateVariable(TemplateNode):
    def __init__(self, var_name: str, escape: bool = True) -> None:
        self.var_name = var_name
        self.escape = escape

    def render(self, template_vars) -> str:
        if self.escape:
            return html.escape(str(template_vars.get(self.var_name) or ""))

        return str(template_vars.get(self.var_name) or "")



from typing import Optional

from graphql.type import GraphQLSchema

from ...objects import ObjectType
from ...types import Resolver


class FederatedObjectType(ObjectType):
    """Add `__referenceResolver` to objects as per apollo-federation."""

    _reference_resolver: Optional[Resolver] = None

    def reference_resolver(self, arg: Optional[Resolver] = None) -> Resolver:
        def register_reference_resolver(f: Resolver) -> Resolver:
            self._reference_resolver = f
            return f

        if callable(arg):
            return register_reference_resolver(arg)
        return register_reference_resolver

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        super().bind_to_schema(schema)

        if callable(self._reference_resolver):
            graphql_type = schema.type_map.get(self.name)
            setattr(
                graphql_type,
                "__resolve_reference__",
                self._reference_resolver,
            )



import asyncio
from collections.abc import Mapping
from functools import wraps
from typing import Optional, Union, Callable, Dict, Any, cast
from warnings import warn

from graphql.language import DocumentNode, OperationDefinitionNode, OperationType
from graphql import GraphQLError, GraphQLType, parse


def convert_camel_case_to_snake(graphql_name: str) -> str:
    """Converts a string with `camelCase` name to `snake_case`.

    Utility function used by Ariadne's name conversion logic for mapping GraphQL
    names using the `camelCase` convention to Python counterparts in `snake_case`.

    Returns a string with converted name.

    # Required arguments

    `graphql_name`: a `str` with name to convert.

    # Example

    All characters in converted name are lowercased:

    ```python
    assert convert_camel_case_to_snake("URL") == "url"
    ```

    `_` is inserted before every uppercase character that's not first and is not
    preceded by other uppercase character:

    ```python
    assert convert_camel_case_to_snake("testURL") == "test_url"
    ```

    `_` is inserted before every uppercase character succeeded by lowercased
    character:

    ```python
    assert convert_camel_case_to_snake("URLTest") == "url_test"
    ```

    `_` is inserted before every digit that's not first and is not preceded by
    other digit:

    ```python
    assert convert_camel_case_to_snake("Rfc123") == "rfc_123"
    ```
    """

    # pylint: disable=too-many-boolean-expressions
    max_index = len(graphql_name) - 1
    lowered_name = graphql_name.lower()

    python_name = ""
    for i, c in enumerate(lowered_name):
        if i > 0 and (
            # testWord -> test_word
            (
                c != graphql_name[i]
                and graphql_name[i - 1] != "_"
                and graphql_name[i - 1] == python_name[-1]
            )
            # TESTWord -> test_word
            or (
                i < max_index
                and graphql_name[i] != c
                and graphql_name[i + 1] == lowered_name[i + 1]
            )
            # test134 -> test_134
            or (c.isdigit() and not graphql_name[i - 1].isdigit())
            # 134test -> 134_test
            or (not c.isdigit() and graphql_name[i - 1].isdigit())
        ):
            python_name += "_"
        python_name += c
    return python_name


def gql(value: str) -> str:
    """Verifies that given string is a valid GraphQL.

    Provides definition time validation for GraphQL strings. Returns unmodified
    string. Some IDEs provide GraphQL syntax for highlighting those strings.


    # Examples

    Python string in this code snippet will use GraphQL's syntax highlighting when
    viewed in VSCode:

    ```python
    type_defs = gql(
        \"\"\"
        type Query {
            hello: String!
        }
        \"\"\"
    )
    ```

    This code will raise a GraphQL parsing error:

    ```python
    type_defs = gql(
        \"\"\"
        type Query {
            hello String!
        }
        \"\"\"
    )
    ```
    """
    parse(value)
    return value


def unwrap_graphql_error(
    error: Union[GraphQLError, Optional[Exception]]
) -> Optional[Exception]:
    """Recursively unwrap exception when its instance of GraphQLError.

    GraphQL query executor wraps exceptions in `GraphQLError` instances which
    contain information about exception's origin in GraphQL query or it's result.

    Original exception is available through `GraphQLError`'s `original_error`
    attribute, but sometimes `GraphQLError` can be wrapped in other `GraphQLError`.

    Returns unwrapped exception or `None` if no original exception was found.

    # Example

    Below code unwraps original `KeyError` from multiple `GraphQLError` instances:

    ```python
    error = KeyError("I am a test!")

    assert unwrap_graphql_error(
        GraphQLError(
            "Error 1",
            GraphQLError(
                "Error 2",
                GraphQLError(
                    "Error 3",
                    original_error=error
                )
            )
        )
    ) == error
    ```

    Passing other exception to `unwrap_graphql_error` results in same exception
    being returned:

    ```python
    error = ValueError("I am a test!")
    assert unwrap_graphql_error(error) == error
    ```
    """

    if isinstance(error, GraphQLError):
        return unwrap_graphql_error(error.original_error)
    return error


def convert_kwargs_to_snake_case(func: Callable) -> Callable:
    """Decorator for resolvers recursively converting their kwargs to `snake_case`.

    Converts keys in `kwargs` dict from `camelCase` to `snake_case` using the
    `convert_camel_case_to_snake` function. Walks values recursively, applying
    same conversion to keys of nested dicts and dicts in lists of elements.

    Returns decorated resolver function.

    > **Deprecated:** This decorator is deprecated and will be deleted in future
    version of Ariadne. Set `out_name`s explicitly in your GraphQL schema or use
    the `convert_schema_names` option on `make_executable_schema`.
    """

    def convert_to_snake_case(m: Mapping) -> Dict:
        converted: Dict = {}
        for k, v in m.items():
            if isinstance(v, Mapping):
                v = convert_to_snake_case(v)
            if isinstance(v, list):
                v = [
                    convert_to_snake_case(i) if isinstance(i, Mapping) else i for i in v
                ]
            converted[convert_camel_case_to_snake(k)] = v
        return converted

    if asyncio.iscoroutinefunction(func):

        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            return await func(*args, **convert_to_snake_case(kwargs))

        return async_wrapper

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return func(*args, **convert_to_snake_case(kwargs))

    return wrapper


def type_implements_interface(interface: str, graphql_type: GraphQLType) -> bool:
    """Test if type definition from GraphQL schema implements an interface.

    Returns `True` if type implements interface and `False` if it doesn't.

    # Required arguments

    `interface`: a `str` with name of interface in GraphQL schema.

    `graphql_type`: a `GraphQLType` interface to test. It may or may not have
    the `interfaces` attribute.
    """

    try:
        return interface in [i.name for i in graphql_type.interfaces]  # type: ignore
    except AttributeError:
        pass

    return False


def get_operation_type(
    graphql_document: DocumentNode, operation_name: Optional[str] = None
) -> OperationType:
    if operation_name:
        for d in graphql_document.definitions:
            d = cast(OperationDefinitionNode, d)
            if d.name and d.name.value == operation_name:
                return d.operation
    else:
        for definition in graphql_document.definitions:
            if isinstance(definition, OperationDefinitionNode):
                return definition.operation
    raise RuntimeError("Can't get GraphQL operation type")


def context_value_one_arg_deprecated():  # TODO: remove in 0.20
    warn(
        "'context_value(request)' has been deprecated and will raise a type "
        "error in Ariadne 0.20. Change definition to "
        "'context_value(request, data)'.",
        DeprecationWarning,
        stacklevel=2,
    )



from typing import Dict, Optional, cast

from graphql import GraphQLInputObjectType, GraphQLSchema
from graphql.type.definition import GraphQLInputFieldOutType, GraphQLNamedType

from .types import SchemaBindable


class InputType(SchemaBindable):
    """Bindable populating input types in a GraphQL schema with Python logic.

    # Example input value represented as dataclass

    Following code creates a GraphQL schema with object type named `Query`
    with single field which has an argument of an input type. It then uses
    the `InputType` to set `ExampleInput` dataclass as Python representation
    of this GraphQL type:

    ```python
    from dataclasses import dataclass

    from ariadne import InputType, QueryType, make_executable_schema

    @dataclass
    class ExampleInput:
        id: str
        message: str

    query_type = QueryType()

    @query_type.field("repr")
    def resolve_repr(*_, input: ExampleInput):
        return repr(input)

    schema = make_executable_schema(
        \"\"\"
        type Query {
            repr(input: ExampleInput): String!
        }

        input ExampleInput {
            id: ID!
            message: String!
        }
        \"\"\",
        query_type,
        # Lambda is used because out type (second argument of InputType)
        # is called with single dict and dataclass requires each value as
        # separate argument.
        InputType("ExampleInput", lambda data: ExampleInput(**data)),
    )
    ```

    # Example input with its fields mapped to custom dict keys

    Following code creates a GraphQL schema with object type named `Query`
    with single field which has an argument of an input type. It then uses
    the `InputType` to set custom "out names" values, mapping GraphQL
    `shortMessage` to `message` key in Python dict:

    ```python
    from ariadne import InputType, QueryType, make_executable_schema

    query_type = QueryType()

    @query_type.field("repr")
    def resolve_repr(*_, input: dict):
        # Dict will have `id` and `message` keys
        input_id = input["id"]
        input_message = input["message"]
        return f"id: {input_id}, message: {input_message}"

    schema = make_executable_schema(
        \"\"\"
        type Query {
            repr(input: ExampleInput): String!
        }

        input ExampleInput {
            id: ID!
            shortMessage: String!
        }
        \"\"\",
        query_type,
        InputType("ExampleInput", out_names={"shortMessage": "message"}),
    )
    ```

    # Example input value as dataclass with custom named fields

    Following code creates a GraphQL schema with object type named `Query`
    with single field which has an argument of an input type. It then uses
    the `InputType` to set `ExampleInput` dataclass as Python representation
    of this GraphQL type, and maps `shortMessage` input field to it's
    `message` attribute:

    ```python
    from dataclasses import dataclass

    from ariadne import InputType, QueryType, make_executable_schema

    @dataclass
    class ExampleInput:
        id: str
        message: str

    query_type = QueryType()

    @query_type.field("repr")
    def resolve_repr(*_, input: ExampleInput):
        return repr(input)

    schema = make_executable_schema(
        \"\"\"
        type Query {
            repr(input: ExampleInput): String!
        }

        input ExampleInput {
            id: ID!
            shortMessage: String!
        }
        \"\"\",
        query_type,
        InputType(
            "ExampleInput",
            lambda data: ExampleInput(**data),
            {"shortMessage": "message"},
        ),
    )
    ```
    """

    _out_type: Optional[GraphQLInputFieldOutType]
    _out_names: Optional[Dict[str, str]]

    def __init__(
        self,
        name: str,
        out_type: Optional[GraphQLInputFieldOutType] = None,
        out_names: Optional[Dict[str, str]] = None,
    ) -> None:
        """Initializes the `InputType` with a `name` and optionally out type
        and out names.

        # Required arguments

        `name`: a `str` with the name of GraphQL object type in GraphQL schema to
        bind to.

        # Optional arguments

        `out_type`: a `GraphQLInputFieldOutType`, Python callable accepting single
        argument, a dict with data from GraphQL query, required to return
        a Python representation of input type.

        `out_names`: a `Dict[str, str]` with mappings from GraphQL field names
        to dict keys in a Python dictionary used to contain a data passed as
        input.
        """
        self.name = name
        self._out_type = out_type
        self._out_names = out_names

    def bind_to_schema(self, schema: GraphQLSchema) -> None:
        """Binds this `InputType` instance to the instance of GraphQL schema.

        if it has an out type function, it assigns it to GraphQL type's
        `out_type` attribute. If type already has other function set on
        it's `out_type` attribute, this type is replaced with new one.

        If it has any out names set, it assigns those to GraphQL type's
        fields `out_name` attributes. If field already has other out name set on
        its `out_name` attribute, this name is replaced with the new one.
        """
        graphql_type = schema.type_map.get(self.name)
        self.validate_graphql_type(graphql_type)
        graphql_type = cast(GraphQLInputObjectType, graphql_type)

        if self._out_type:
            graphql_type.out_type = self._out_type  # type: ignore

        if self._out_names:
            for graphql_name, python_name in self._out_names.items():
                if graphql_name not in graphql_type.fields:
                    raise ValueError(
                        f"Field {graphql_name} is not defined on type {self.name}"
                    )
                graphql_type.fields[graphql_name].out_name = python_name

    def validate_graphql_type(self, graphql_type: Optional[GraphQLNamedType]) -> None:
        """Validates that schema's GraphQL type associated with this `InputType`
        is an `input`."""
        if not graphql_type:
            raise ValueError("Type %s is not defined in the schema" % self.name)
        if not isinstance(graphql_type, GraphQLInputObjectType):
            raise ValueError(
                "%s is defined in the schema, but it is instance of %s (expected %s)"
                % (
                    self.name,
                    type(graphql_type).__name__,
                    GraphQLInputObjectType.__name__,
                )
            )



import asyncio
from inspect import isawaitable
from typing import Any, AsyncGenerator, Dict, List, Optional, cast

from graphql import DocumentNode, GraphQLError
from graphql.language import OperationType
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState

from ...graphql import parse_query, subscribe, validate_data
from ...logger import log_error
from ...types import (
    Operation,
    WebSocketConnectionError,
)
from ...utils import get_operation_type
from .base import GraphQLWebsocketHandler

# Note: Confusingly, the subscriptions-transport-ws library
# calls its WebSocket subprotocol graphql-ws,
# and the graphql-ws library calls its subprotocol graphql-transport-ws!


class GraphQLWSHandler(GraphQLWebsocketHandler):
    """Implementation of the (older) graphql-ws subprotocol from the
    subscriptions-transport-ws library.

    For more details see it's GH page:

    https://github.com/apollographql/subscriptions-transport-ws/blob/master/PROTOCOL.md
    """

    keepalive: Optional[float]

    GQL_CONNECTION_INIT = "connection_init"  # Client -> Server
    GQL_CONNECTION_ACK = "connection_ack"  # Server -> Client
    GQL_CONNECTION_ERROR = "connection_error"  # Server -> Client

    # NOTE: The keep alive message type does not follow the standard due to connection optimizations
    GQL_CONNECTION_KEEP_ALIVE = "ka"  # Server -> Client

    GQL_CONNECTION_TERMINATE = "connection_terminate"  # Client -> Server
    GQL_START = "start"  # Client -> Server
    GQL_DATA = "data"  # Server -> Client
    GQL_ERROR = "error"  # Server -> Client
    GQL_COMPLETE = "complete"  # Server -> Client
    GQL_STOP = "stop"  # Client -> Server

    def __init__(
        self,
        *args,
        keepalive: Optional[float] = None,
        **kwargs,
    ) -> None:
        """Initializes the websocket handler.

        # Optional arguments

        `keepalive`: a `float` with time frequency for sending the keep-alive
        messages to clients with open websocket connections. `1.0` is 1 second.
        If set to `None` or `0`, no keep-alive messages are sent.
        Defaults to `None`.
        """

        super().__init__(*args, **kwargs)
        self.keepalive = keepalive

    async def handle(self, scope: Scope, receive: Receive, send: Send):
        """An entrypoint for the GraphQL WebSocket handler.

        This method is called by the Ariadne ASGI GraphQL application to handle
        the websocket connections.

        It creates the `starlette.websockets.WebSocket` instance and calls
        `handle_websocket` method with it.

        # Required arguments

        `scope`: The connection scope information, a dictionary that contains
        at least a type key specifying the protocol that is incoming.

        `receive`: an awaitable callable that will yield a new event dictionary
        when one is available.

        `send`: an awaitable callable taking a single event dictionary as a
        positional argument that will return once the send has been completed
        or the connection has been closed.

        Details about the arguments and their usage are described in the
        ASGI specification:

        https://asgi.readthedocs.io/en/latest/specs/main.html
        """
        websocket = WebSocket(scope=scope, receive=receive, send=send)
        await self.handle_websocket(websocket)

    async def handle_websocket(self, websocket: WebSocket):
        """Handle GraphQL the WebSocket connection.

        Is called by the `handle` method and `handle_websocket` method of the
        ASGI GraphQL application.

        # Required arguments:

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.
        """
        operations: Dict[str, Operation] = {}
        await websocket.accept("graphql-ws")
        try:
            while WebSocketState.DISCONNECTED not in (
                websocket.client_state,
                websocket.application_state,
            ):
                message = await websocket.receive_json()
                await self.handle_websocket_message(websocket, message, operations)
        except WebSocketDisconnect:
            pass
        finally:
            for operation_id, operation in list(operations.items()):
                await self.stop_websocket_operation(websocket, operation)
                del operations[operation_id]

            try:
                if self.on_disconnect:
                    result = self.on_disconnect(websocket)
                    if result and isawaitable(result):
                        await result
            except Exception as error:
                if not isinstance(error, GraphQLError):
                    error = GraphQLError(str(error), original_error=error)
                log_error(error, self.logger)

    async def handle_websocket_message(
        self,
        websocket: WebSocket,
        message: dict,
        operations: Dict[str, Operation],
    ):
        """Handles new message from websocket connection.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `message`: a `dict` with message payload.

        `operations`: a `dict` with currently active GraphQL operations.
        """
        operation_id = cast(str, message.get("id"))
        message_type = cast(str, message.get("type"))

        if message_type == GraphQLWSHandler.GQL_CONNECTION_INIT:
            await self.handle_websocket_connection_init_message(websocket, message)
        elif message_type == GraphQLWSHandler.GQL_CONNECTION_TERMINATE:
            await self.handle_websocket_connection_terminate_message(websocket)
        elif message_type == GraphQLWSHandler.GQL_START:
            await self.process_single_message(
                websocket, message.get("payload"), operation_id, operations
            )
        elif message_type == GraphQLWSHandler.GQL_STOP:
            if operation_id in operations:
                await self.stop_websocket_operation(websocket, operations[operation_id])
                del operations[operation_id]

    async def process_single_message(
        self,
        websocket: WebSocket,
        data: Any,
        operation_id: str,
        operations: Dict[str, Operation],
    ) -> None:
        """Processes websocket message containing new GraphQL operation.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `data`: a `dict` with message's payload.

        `operation_id`: a `str` with an ID of new operation.

        `operations`: a `dict` with currently active GraphQL operations.
        """

        validate_data(data)
        context_value = await self.get_context_for_request(websocket, data)

        try:
            query_document = parse_query(context_value, self.query_parser, data)
        except GraphQLError as error:
            log_error(error, self.logger)
            await websocket.send_json(
                {
                    "type": GraphQLWSHandler.GQL_ERROR,
                    "id": operation_id,
                    "payload": self.error_formatter(error, self.debug),
                }
            )
            return

        operation_type = get_operation_type(query_document, data.get("operationName"))

        if operation_type == OperationType.SUBSCRIPTION:
            await self.start_websocket_operation(
                websocket,
                data,
                context_value,
                query_document,
                operation_id,
                operations,
            )
        else:
            if self.http_handler is None:
                raise TypeError(
                    "http_handler is not set, call configure method to initialize it"
                )
            _, result = await self.http_handler.execute_graphql_query(websocket, data)
            await websocket.send_json(
                {
                    "type": GraphQLWSHandler.GQL_DATA,
                    "id": operation_id,
                    "payload": result,
                }
            )

    async def handle_websocket_connection_init_message(
        self, websocket: WebSocket, message: dict
    ):
        """Handles `connection_init` websocket message.

        Initializes new websocket instance.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `message`: a `dict` with message's payload.
        """
        try:
            if self.on_connect:
                result = self.on_connect(websocket, message.get("payload"))
                if result and isawaitable(result):
                    await result

            await websocket.send_json({"type": GraphQLWSHandler.GQL_CONNECTION_ACK})
            asyncio.ensure_future(self.keep_websocket_alive(websocket))
        except Exception as error:
            log_error(error, self.logger)

            if isinstance(error, WebSocketConnectionError):
                payload = error.payload  # pylint: disable=no-member
            else:
                payload = {"message": "Unexpected error has occurred."}

            await websocket.send_json(
                {"type": GraphQLWSHandler.GQL_CONNECTION_ERROR, "payload": payload}
            )
            await websocket.close()

    async def handle_websocket_connection_terminate_message(
        self,
        websocket: WebSocket,
    ):
        """Handles `terminate` websocket message.

        Closes open websocket connection.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.
        """
        await websocket.close()

    async def keep_websocket_alive(self, websocket: WebSocket):
        if not self.keepalive:
            return
        while websocket.application_state != WebSocketState.DISCONNECTED:
            try:
                await websocket.send_json(
                    {"type": GraphQLWSHandler.GQL_CONNECTION_KEEP_ALIVE}
                )
            except WebSocketDisconnect:
                return
            await asyncio.sleep(self.keepalive)

    async def start_websocket_operation(
        self,
        websocket: WebSocket,
        data: Any,
        context_value: Any,
        query_document: DocumentNode,
        operation_id: str,
        operations: Dict[str, Operation],
    ):
        if self.schema is None:
            raise TypeError("schema is not set, call configure method to initialize it")

        success, results = await subscribe(
            self.schema,
            data,
            context_value=context_value,
            root_value=self.root_value,
            query_document=query_document,
            validation_rules=self.validation_rules,
            debug=self.debug,
            introspection=self.introspection,
            logger=self.logger,
            error_formatter=self.error_formatter,
        )

        if not success:
            results = cast(List[dict], results)
            await websocket.send_json(
                {
                    "type": GraphQLWSHandler.GQL_ERROR,
                    "id": operation_id,
                    "payload": results[0],
                }
            )
        else:
            results = cast(AsyncGenerator, results)
            operations[operation_id] = Operation(
                id=operation_id,
                name=data.get("operationName"),
                generator=results,
            )

            if self.on_operation:
                try:
                    result = self.on_operation(websocket, operations[operation_id])
                    if result and isawaitable(result):
                        await result
                except Exception as error:
                    if not isinstance(error, GraphQLError):
                        error = GraphQLError(str(error), original_error=error)
                    log_error(error, self.logger)

            asyncio.ensure_future(
                self.observe_async_results(websocket, results, operation_id)
            )

    async def stop_websocket_operation(
        self,
        websocket: WebSocket,
        operation: Operation,
    ):
        """Stops specified GraphQL operation for given connection.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `operation`: an `Operation` to stop.
        """
        if self.on_complete:
            try:
                result = self.on_complete(websocket, operation)
                if result and isawaitable(result):
                    await result
            except Exception as error:
                if not isinstance(error, GraphQLError):
                    error = GraphQLError(str(error), original_error=error)
                log_error(error, self.logger)

        await operation.generator.aclose()

    async def observe_async_results(
        self,
        websocket: WebSocket,
        results: AsyncGenerator,
        operation_id: str,
    ) -> None:
        """Converts results from Ariadne's `subscribe` generator into websocket
        messages it next sends to the client.

        # Required arguments

        `websocket`: the `WebSocket` instance from Starlette or FastAPI.

        `results`: the `AsyncGenerator` returned from Ariadne's
        `subscribe` function.

        `operation_id`: a `str` with id of operation.
        """
        try:
            async for result in results:
                payload = {}
                if result.data:
                    payload["data"] = result.data
                if result.errors:
                    for error in result.errors:
                        log_error(error, self.logger)
                    payload["errors"] = [
                        self.error_formatter(error, self.debug)
                        for error in result.errors
                    ]
                await websocket.send_json(
                    {
                        "type": GraphQLWSHandler.GQL_DATA,
                        "id": operation_id,
                        "payload": payload,
                    }
                )
        except Exception as error:
            if not isinstance(error, GraphQLError):
                error = GraphQLError(str(error), original_error=error)
            log_error(error, self.logger)
            payload = {"errors": [self.error_formatter(error, self.debug)]}
            await websocket.send_json(
                {
                    "type": GraphQLWSHandler.GQL_DATA,
                    "id": operation_id,
                    "payload": payload,
                }
            )

        if WebSocketState.DISCONNECTED not in (
            websocket.client_state,
            websocket.application_state,
        ):
            await websocket.send_json(
                {"type": GraphQLWSHandler.GQL_COMPLETE, "id": operation_id}
            )



from typing import Optional, Union
from typing_extensions import Protocol

from .exceptions import HttpBadRequestError
from .scalars import ScalarType

SPEC_URL = "https://github.com/jaydenseric/graphql-multipart-request-spec"


class FilesDict(Protocol):
    def __getitem__(self, key):
        ...  # pragma: no-cover


def combine_multipart_data(
    operations: Union[dict, list], files_map: dict, files: FilesDict
) -> Union[dict, list]:
    """Populates `operations` variables with `files` using the `files_map`.

    Utility function for integration developers.

    Mutates `operations` in place, but also returns it.

    # Requires arguments

    `operations`: a `list` or `dict` with GraphQL operations to populate the file
    variables in. It contains `operationName`, `query` and `variables` keys, but
    implementation only cares about `variables` being present.

    `files_map`: a `dict` with mapping of `files` to `operations`. Keys correspond
    to keys in `files dict`, values are lists of strings with paths (eg.:
    `variables.key.0` maps to `operations["variables"]["key"]["0"]`).

    `files`: a `dict` of files. Keys are strings, values are environment specific
    representations of uploaded files.

    # Example

    Following example uses `combine_multipart_data` to populate the `image`
    variable with file object from `files`, using the `files_map` to know
    which variable to replace.

    ```python
    # Single GraphQL operation
    operations = {
        "operationName": "AvatarUpload",
        "query": \"\"\"
            mutation AvatarUpload($type: String!, $image: Upload!) {
                avatarUpload(type: $type, image: $image) {
                    success
                    errors
                }
            }
        \"\"\",
        "variables": {"type": "SQUARE", "image": None}
    }
    files_map = {"0": ["variables.image"]}
    files = {"0": UploadedFile(....)}

    combine_multipart_data(operations, files_map, files

    assert operations == {
        "variables": {"type": "SQUARE", "image": UploadedFile(....)}
    }
    ```
    """

    if not isinstance(operations, (dict, list)):
        raise HttpBadRequestError(
            "Invalid type for the 'operations' multipart field ({}).".format(SPEC_URL)
        )
    if not isinstance(files_map, dict):
        raise HttpBadRequestError(
            "Invalid type for the 'map' multipart field ({}).".format(SPEC_URL)
        )

    files_map = inverse_files_map(files_map, files)
    if isinstance(operations, list):
        for i, operation in enumerate(operations):
            add_files_to_variables(
                operation.get("variables"), "{}.variables".format(i), files_map
            )
    if isinstance(operations, dict):
        add_files_to_variables(operations.get("variables"), "variables", files_map)
    return operations


def inverse_files_map(files_map: dict, files: FilesDict) -> dict:
    inverted_map = {}
    for field_name, paths in files_map.items():
        if not isinstance(paths, list):
            raise HttpBadRequestError(
                (
                    "Invalid type for the 'map' multipart field entry "
                    "key '{}' array ({})."
                ).format(field_name, SPEC_URL)
            )

        for i, path in enumerate(paths):
            if not isinstance(path, str):
                raise HttpBadRequestError(
                    (
                        "Invalid type for the 'map' multipart field entry key "
                        "'{}' array index '{}' value ({})."
                    ).format(field_name, i, SPEC_URL)
                )

            try:
                inverted_map[path] = files[field_name]
            except KeyError as ex:
                raise HttpBadRequestError(
                    ("File data was missing for entry key '{}' ({}).").format(
                        field_name, SPEC_URL
                    )
                ) from ex

    return inverted_map


def add_files_to_variables(
    variables: Optional[Union[dict, list]], path: str, files_map: dict
):
    if isinstance(variables, dict):
        for variable, value in variables.items():
            variable_path = "{}.{}".format(path, variable)
            if isinstance(value, (dict, list)):
                add_files_to_variables(value, variable_path, files_map)
            elif value is None:
                variables[variable] = files_map.get(variable_path)

    if isinstance(variables, list):
        for i, value in enumerate(variables):
            variable_path = "{}.{}".format(path, i)
            if isinstance(value, (dict, list)):
                add_files_to_variables(value, variable_path, files_map)
            elif value is None:
                variables[i] = files_map.get(variable_path)


"""Optional Python logic for `Upload` scalar.

`Upload` scalar doesn't require any custom Python logic to work, but this utility 
sets `serializer` and `literal_parser` to raise ValueErrors when `Upload` is used 
either as return type for field or passed as literal value in GraphQL query.

# Example

Below code defines a schema with `Upload` scalar using `upload_scalar` utility:

```python
from ariadne import MutationType, make_executable_schema, upload_scalar

mutation_type = MutationType()

@mutation_type.field("handleUpload")
def resolve_handle_upload(*_, upload):
    return repr(upload)


schema = make_executable_schema(
    \"\"\"
    scalar Upload

    type Query {
        empty: String
    }

    type Mutation {
        handleUpload(upload: Upload!): String
    }
    \"\"\",
    upload_scalar,
    mutation_type,
)
```
"""
upload_scalar = ScalarType("Upload")


@upload_scalar.serializer
def serialize_upload(*_):
    raise ValueError("'Upload' scalar serialization is not supported.")


@upload_scalar.literal_parser
def parse_upload_literal(*_):
    raise ValueError("'Upload' scalar literal is not supported.")


@upload_scalar.value_parser
def parse_upload_value(value):
    return value



import re
import os
from typing import Dict, List, Optional, Type, Union, cast

from graphql import extend_schema, parse
from graphql.language import DocumentNode
from graphql.language.ast import ObjectTypeDefinitionNode
from graphql.type import (
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLUnionType,
)

from ...executable_schema import (
    SchemaBindables,
    make_executable_schema,
    join_type_defs,
)
from ...schema_names import SchemaNameConverter
from ...schema_visitor import SchemaDirectiveVisitor
from ...load_schema import load_schema_from_path
from .utils import get_entity_types, purge_schema_directives, resolve_entities

base_federation_service_type_defs = """
    scalar _Any

    type _Service {{
        sdl: String
    }}

    {type_token} Query {{
        _service: _Service!
    }}
"""

federation_entity_type_defs = """
    union _Entity

    extend type Query {
        _entities(representations: [_Any!]!): [_Entity]!
    }
"""


def has_query_type(type_defs: str) -> bool:
    ast_document = parse(type_defs)
    return any(
        (
            isinstance(definition, ObjectTypeDefinitionNode)
            and definition.name.value == "Query"
        )
        for definition in ast_document.definitions
    )


def make_federated_schema(
    type_defs: Union[str, List[str]],
    *bindables: SchemaBindables,
    directives: Optional[Dict[str, Type[SchemaDirectiveVisitor]]] = None,
    convert_names_case: Union[bool, SchemaNameConverter] = False,
) -> GraphQLSchema:
    if isinstance(type_defs, list):
        type_defs = join_type_defs(type_defs)

    # Remove custom schema directives (to avoid apollo-gateway crashes).
    # NOTE: This does NOT interfere with ariadne's directives support.
    sdl = purge_schema_directives(type_defs)
    type_token = "extend type" if has_query_type(sdl) else "type"
    federation_service_type = base_federation_service_type_defs.format(
        type_token=type_token
    )

    tdl = [type_defs, federation_service_type]

    link = re.search(
        r"(?<=@link).*?url:.*?\"(.*?)\".?[^)]+?", sdl, re.MULTILINE | re.DOTALL
    )  # use regex to parse if it's fed 1 or fed 2; adds dedicated typedefs per spec
    dirname = os.path.dirname(__file__)
    if link and link.group(1) == "https://specs.apollo.dev/federation/v2.0":
        definitions = load_schema_from_path(
            os.path.join(dirname, "./definitions/fed2_0.graphql")
        )
    elif link and link.group(1) == "https://specs.apollo.dev/federation/v2.1":
        definitions = load_schema_from_path(
            os.path.join(dirname, "./definitions/fed2_1.graphql")
        )
    elif link and link.group(1) == "https://specs.apollo.dev/federation/v2.2":
        definitions = load_schema_from_path(
            os.path.join(dirname, "./definitions/fed2_2.graphql")
        )
    elif link and link.group(1) == "https://specs.apollo.dev/federation/v2.3":
        definitions = load_schema_from_path(
            os.path.join(dirname, "./definitions/fed2_3.graphql")
        )
    else:
        definitions = load_schema_from_path(
            os.path.join(dirname, "./definitions/fed1_0.graphql")
        )

    tdl.append(definitions)

    type_defs = join_type_defs(tdl)
    schema = make_executable_schema(
        type_defs,
        *bindables,
        directives=directives,
        convert_names_case=convert_names_case,
    )

    # Parse through the schema to find all entities with key directive.
    entity_types = get_entity_types(schema)
    has_entities = len(entity_types) > 0

    # Add the federation type definitions.
    if has_entities:
        schema = extend_federated_schema(schema, parse(federation_entity_type_defs))

        # Add _entities query.
        entity_type = schema.get_type("_Entity")
        if entity_type:
            entity_type = cast(GraphQLUnionType, entity_type)
            entity_type.types = entity_types

        query_type = schema.get_type("Query")
        if query_type:
            query_type = cast(GraphQLObjectType, query_type)
            query_type.fields["_entities"].resolve = resolve_entities

    # Add _service query.
    query_type = schema.get_type("Query")
    if query_type:
        query_type = cast(GraphQLObjectType, query_type)
        query_type.fields["_service"].resolve = lambda _service, info: {"sdl": sdl}

    return schema


def extend_federated_schema(
    schema: GraphQLSchema,
    document_ast: DocumentNode,
    assume_valid: bool = False,
    assume_valid_sdl: bool = False,
) -> GraphQLSchema:
    extended_schema = extend_schema(
        schema,
        document_ast,
        assume_valid,
        assume_valid_sdl,
    )

    for k, v in schema.type_map.items():
        resolve_reference = getattr(v, "__resolve_reference__", None)
        if resolve_reference and k in extended_schema.type_map:
            setattr(
                extended_schema.type_map[k],
                "__resolve_reference__",
                resolve_reference,
            )

    return extended_schema
